@@ -6,6 +6,8 @@
 
 #include "pg_statsinfod.h"
 
+#include "libpq/pqsignal.h"
+
 #include <dirent.h>
 #include <fcntl.h>
 #include <sys/stat.h>
@@ -13,6 +15,7 @@
 
 #ifndef WIN32
 #include <dirent.h>
+#include <linux/version.h>
 #endif
 
 /* pg_statsinfo.control values */
@@ -41,6 +44,7 @@ static char		   *my_adjust_log_error;
 static char		   *my_adjust_log_log;
 static char		   *my_adjust_log_fatal;
 static List		   *my_textlog_nologging_users;
+static int		    my_controlfile_fsync_interval;
 /*-----------------------*/
 
 #if PG_VERSION_NUM < 90000
@@ -97,7 +101,7 @@ static void logger_parse(Logger *logger, const char *pg_log, bool only_routing);
 static void logger_route(Logger *logger, const Log *log);
 
 static void init_log(Log *log, const char *buf, size_t len, const size_t fields[]);
-static bool logger_open(Logger *logger, const char *csvlog);
+static bool logger_open(Logger *logger, const char *csvlog, long offset);
 static void logger_close(Logger *logger);
 static bool logger_next(Logger *logger, const char *pg_log);
 static void get_csvlog(char csvlog[], const char *prev, const char *pg_log);
@@ -114,8 +118,23 @@ static void open_controlfile(Logger *logger, int flags, mode_t mode);
 static void load_controlfile(Logger *logger);
 static bool ReadControlFile(void);
 static void RewriteControlFile(Logger *logger);
+static void FsyncControlFile(void);
+static void FlushControlFile(void);
+static int SyncFile(int fd);
+static time_t get_next_time(time_t now, int interval);
 static void logger_shutdown(Logger *logger);
 
+static volatile bool	stop_request = false;
+
+/* SIGUSR1: instructs to stop the pg_statsinfod process */
+static void
+logger_sigusr1_handler(SIGNAL_ARGS)
+{
+	stop_request = true;
+	if (shutdown_state < SHUTDOWN_REQUESTED)
+		shutdown_progress(SHUTDOWN_REQUESTED);
+}
+
 void
 logger_init(void)
 {
@@ -134,11 +153,19 @@ logger_main(void *arg)
 	Logger			  logger;
 	int				  entry;
 	struct dirent	**dplist;
+	time_t			  now;
+	time_t			  next_fsync;
+
+	/* Set up signal handlers */
+	pqsignal(SIGUSR1, logger_sigusr1_handler);
 
 	memset(&logger, 0, sizeof(logger));
 	initStringInfo(&logger.buf);
 	reload_params();
 
+	now = time(NULL);
+	next_fsync = get_next_time(now, my_controlfile_fsync_interval);
+
 	/* sets latest csvlog to the elements of logger */
 	for (;;)
 	{
@@ -159,16 +186,20 @@ logger_main(void *arg)
 	ControlFile.state = STATSINFO_RUNNING;
 
 	/* perform the log routing until end of the latest csvlog */
-	logger_open(&logger, logger.csv_path);
+	logger_open(&logger, logger.csv_path, logger.csv_offset);
 	logger_parse(&logger, my_log_directory, true);
 
+	FsyncControlFile();
+
 	/*
 	 * Logger should not shutdown before any other threads are alive,
 	 * or postmaster exists unless shutdown log is not found.
 	 */
 	while (shutdown_state < WRITER_SHUTDOWN ||
-		   (!shutdown_message_found && postmaster_is_alive()))
+		   (!shutdown_message_found && postmaster_is_alive() && !stop_request))
 	{
+		now = time(NULL);
+
 		/* update settings if reloaded */
 		if (logger_reload_time < collector_reload_time)
 		{
@@ -184,11 +215,25 @@ logger_main(void *arg)
 			}
 			else
 				assign_textlog_path(&logger, my_log_directory);
+
+			if (my_controlfile_fsync_interval >= 0)
+			{
+				now = time(NULL);
+				next_fsync = get_next_time(now, my_controlfile_fsync_interval);
+			}
 		}
 
 		logger_parse(&logger, my_log_directory, false);
 		usleep(200 * 1000);	/* 200ms */
 
+		/* fsync control file */
+		if (my_controlfile_fsync_interval >= 0 && now >= next_fsync)
+		{
+			FlushControlFile();
+			now = time(NULL);
+			next_fsync = get_next_time(now, my_controlfile_fsync_interval);
+		}
+
 		/* check postmaster pid. */
 		if (shutdown_state < SHUTDOWN_REQUESTED && !postmaster_is_alive())
 			shutdown_progress(SHUTDOWN_REQUESTED);
@@ -197,7 +242,7 @@ logger_main(void *arg)
 	}
 
 	/* exit after some delay */
-	if (!shutdown_message_found)
+	if (!shutdown_message_found && !stop_request)
 	{
 		time_t	until = time(NULL) + LOGGER_EXIT_DELAY;
 
@@ -211,17 +256,23 @@ logger_main(void *arg)
 		}
 	}
 
+	/* update pg_statsinfo.control */
+	if (shutdown_message_found || !stop_request)
+		ControlFile.state = STATSINFO_SHUTDOWNED;
+	else
+		ControlFile.state = STATSINFO_STOPPED;
+	RewriteControlFile(&logger);
+	FsyncControlFile();
+
 	/* final shutdown message */
 	if (shutdown_message_found)
 		elog(LOG, "shutdown");
+	else if (stop_request)
+		elog(LOG, "shutdown by stop request");
 	else
 		elog(WARNING, "shutdown because server process exited abnormally");
 	logger_recv(&logger);
 
-	/* update pg_statsinfo.control */
-	ControlFile.state = STATSINFO_SHUTDOWNED;
-	RewriteControlFile(&logger);
-
 	logger_close(&logger);
 	shutdown_progress(LOGGER_SHUTDOWN);
 
@@ -452,6 +503,9 @@ reload_params(void)
 		adjust_log_list = add_adlog(adjust_log_list, INFO, my_adjust_log_info);
 	}
 
+	/* fsync interval of control file */
+	my_controlfile_fsync_interval = controlfile_fsync_interval;
+
 	pgut_log_level = Min(textlog_min_messages, syslog_min_messages);
 
 	pthread_mutex_unlock(&reload_lock);
@@ -576,25 +630,9 @@ logger_parse(Logger *logger, const char *pg_log, bool only_routing)
 
 		init_log(&log, logger->buf.data, logger->buf.len, logger->fields);
 
-		/* parse performance logs; those messages are NOT routed. */
+		/* parse operation request logs; those messages are NOT routed. */
 		if (log.elevel == LOG)
 		{
-			/* checkpoint ? */
-			if (is_checkpoint(log.message))
-			{
-				if (!only_routing)
-					parse_checkpoint(log.message, log.timestamp);
-				continue;
-			}
-
-			/* autovacuum ? */
-			if (is_autovacuum(log.message))
-			{
-				if (!only_routing)
-					parse_autovacuum(log.message, log.timestamp);
-				continue;
-			}
-
 			/* snapshot requested ? */
 			if (strcmp(log.message, LOGMSG_SNAPSHOT) == 0)
 			{
@@ -661,6 +699,20 @@ logger_parse(Logger *logger, const char *pg_log, bool only_routing)
 
 		if (!only_routing && save_elevel == LOG)
 		{
+			/* checkpoint ? */
+			if (is_checkpoint(log.message))
+			{
+				parse_checkpoint(log.message, log.timestamp);
+				continue;
+			}
+
+			/* autovacuum ? */
+			if (is_autovacuum(log.message))
+			{
+				parse_autovacuum(log.message, log.timestamp);
+				continue;
+			}
+
 			/* setting parameters reloaded ? */
 			if (strcmp(log.message, msg_sighup) == 0)
 			{
@@ -688,14 +740,14 @@ logger_parse(Logger *logger, const char *pg_log, bool only_routing)
 }
 
 static bool
-logger_open(Logger *logger, const char *csvlog)
+logger_open(Logger *logger, const char *csvlog, long offset)
 {
 	mode_t	mask;
 
 	Assert(!logger->fp);
 	Assert(!logger->textlog);
 
-	assign_csvlog_path(logger, my_log_directory, csvlog, 0);
+	assign_csvlog_path(logger, my_log_directory, csvlog, offset);
 	assign_textlog_path(logger, my_log_directory);
 
 	/* open csvlog file */
@@ -794,9 +846,19 @@ logger_next(Logger *logger, const char *pg_log)
 		 }
 
 		logger_close(logger);
-		if (!logger_open(logger, csvlog))
+		if (!logger_open(logger, csvlog, 0))
 			return false;
 
+		/*
+		 * if writer thread working in fallback mode,
+		 * write in the textlog that agent is working in fallback mode.
+		 */
+		if (writer_state == WRITER_FALLBACK)
+		{
+			elog(LOG, "*** pg_statsinfo is working in fallback mode ***");
+			logger_recv(logger);
+		}
+
 		elog(DEBUG2, "read csvlog \"%s\"", logger->csv_path);
 	}
 
@@ -1066,16 +1128,23 @@ static int
 csvfilter(const struct dirent *dp)
 {
 	const char	*extension;
-
-	if (dp->d_type != DT_REG)
-		return 0;
+	struct stat	 entry_stat;
+	char		 entry_path[MAXPGPATH];
 
 	/* check the extension is .csv */
 	extension = strrchr(dp->d_name, '.');
 	if (extension && strcmp(extension, ".csv") == 0)
-		return 1;
-	else
-		return 0;
+	{
+		/* do lstat() for matching extensions */
+		join_path_components(entry_path, my_log_directory, dp->d_name);
+		if (lstat(entry_path, &entry_stat) == 0)
+		{
+			/* is this a regular file? */
+			if (S_ISREG(entry_stat.st_mode))
+				return 1;
+		}
+	}
+	return 0;
 }
 
 static void
@@ -1112,9 +1181,10 @@ load_controlfile(Logger *logger)
 
 	/*
 	 * if state value of pg_statsinfo.control is not "STATSINFO_SHUTDOWNED",
-	 * that means the previous pg_statsinfod was abnormally end.
-	 * in that case, parse the csvlog between csvlog of the point of abnormal
-	 * termination and latest csvlog.
+	 * that means the previous pg_statsinfod was abnormally end or stopped by
+	 * request.
+	 * in this case, parse the csvlog between last parsed position and
+	 * latest position.
 	 */
 	if (ReadControlFile())
 	{
@@ -1126,11 +1196,29 @@ load_controlfile(Logger *logger)
 		join_path_components(
 			prev_csvlog, my_log_directory, ControlFile.csv_name);
 
+		/*
+		 * if state of pg_statsinfo.control is "STATSINFO_STOPPED",
+		 * restore the latest textlog from previous textlog that was renamed.
+		 */
+		if (ControlFile.state == STATSINFO_STOPPED)
+		{
+			char prev_textlog[MAXPGPATH];
+
+			strlcpy(prev_textlog, prev_csvlog, MAXPGPATH);
+			replace_extension(prev_textlog, ".log");
+
+			if (rename(prev_textlog, logger->textlog_path) != 0)
+				ereport(ERROR,
+					(errcode_errno(),
+					 errmsg("could not rename file \"%s\" to \"%s\": %m",
+					 	prev_textlog, logger->textlog_path)));
+		}
+
 		if (stat(prev_csvlog, &st) == 0 && ControlFile.csv_offset <= st.st_size)
 		{
 			/* set the csvlog path and the csvlog offset to the logger */
 			assign_csvlog_path(logger, my_log_directory,
-			ControlFile.csv_name, ControlFile.csv_offset);
+				ControlFile.csv_name, ControlFile.csv_offset);
 			return;
 		}
 
@@ -1264,6 +1352,56 @@ RewriteControlFile(Logger *logger)
 	}
 }
 
+static void
+FsyncControlFile(void)
+{
+	Assert(cf_fd >= 0);	/* have not been opened the pg_statsinfo.control */
+
+	elog(DEBUG2, "fsync control file (fsync method=\"fsync\")");
+	if (fsync(cf_fd) != 0)
+		ereport(ERROR,
+			(errcode_errno(),
+				errmsg("could not fsync control file \"%s\": %m",
+				STATSINFO_CONTROL_FILE)));
+}
+
+static void
+FlushControlFile(void)
+{
+	Assert(cf_fd >= 0);	/* have not been opened the pg_statsinfo.control */
+
+	if (SyncFile(cf_fd) != 0)
+		ereport(ERROR,
+			(errcode_errno(),
+				errmsg("could not fsync control file \"%s\": %m",
+				STATSINFO_CONTROL_FILE)));
+}
+
+static int
+SyncFile(int fd)
+{
+#if defined(HAVE_SYNC_FILE_RANGE)
+	elog(DEBUG2, "fsync file (fsync method=\"sync_file_range\")");
+	return sync_file_range(fd, 0, 0,
+		SYNC_FILE_RANGE_WAIT_BEFORE | SYNC_FILE_RANGE_WRITE);
+#elif defined(HAVE_FDATASYNC)
+	elog(DEBUG2, "fsync file (fsync method=\"fdatasync\")");
+	return fdatasync(fd);
+#else
+	elog(DEBUG2, "fsync file (fsync method=\"fsync\")");
+	return fsync(cf_fd);
+#endif
+}
+
+static time_t
+get_next_time(time_t now, int interval)
+{
+	if (interval > 0)
+		return (now / interval) * interval + interval;
+	else
+		return now;
+}
+
 static void
 logger_shutdown(Logger *logger)
 {
