@@ -6,11 +6,9 @@
 
 #include "pg_statsinfod.h"
 
-#include <signal.h>
+#include <fcntl.h>
 
-#include "miscadmin.h"
-
-const char *PROGRAM_VERSION	= "2.4.1";
+const char *PROGRAM_VERSION	= "2.5.0";
 const char *PROGRAM_URL		= "http://pgstatsinfo.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 
@@ -57,6 +55,7 @@ char		   *adjust_log_error;
 char		   *adjust_log_log;
 char		   *adjust_log_fatal;
 char		   *textlog_nologging_users;
+int				controlfile_fsync_interval;
 /*---- GUC variables (writer) ----------*/
 char		   *repository_server;
 int			    enable_maintenance;
@@ -81,7 +80,7 @@ char		   *msg_autovacuum;
 char		   *msg_autoanalyze;
 char		   *msg_checkpoint_starting;
 char		   *msg_checkpoint_complete;
-size_t			checkpoint_starting_prefix_len;
+char		   *msg_restartpoint_complete;
 /*--------------------------------------*/
 
 /* current shutdown state */
@@ -89,13 +88,15 @@ volatile ShutdownState	shutdown_state;
 bool					shutdown_message_found;
 static pthread_mutex_t	shutdown_state_lock;
 
+/* current writer state */
+volatile WriterState	writer_state;
+
 /* threads */
 pthread_t	th_collector;
 pthread_t	th_logger;
 pthread_t	th_writer;
 
 static int help(void);
-static bool ensure_schema(PGconn *conn, const char *schema);
 static bool assign_int(const char *value, void *var);
 static bool assign_elevel(const char *value, void *var);
 static bool assign_syslog(const char *value, void *var);
@@ -104,11 +105,11 @@ static bool assign_bool(const char *value, void *var);
 static bool assign_time(const char *value, void *var);
 static bool connect_test(const char *conninfo);
 static void readopt(void);
-static void after_readopt(void);
 static bool decode_time(const char *field, int *hour, int *min, int *sec);
 static int strtoi(const char *nptr, char **endptr, int base);
 static bool execute_script(PGconn *conn, const char *script_file);
-static bool check_repository(PGconn *conn);
+static void create_lock_file(void);
+static void unlink_lock_file(void);
 
 /* parameters */
 static struct ParamMap PARAM_MAP[] =
@@ -151,6 +152,7 @@ static struct ParamMap PARAM_MAP[] =
 	{GUC_PREFIX ".maintenance_time", assign_time, &maintenance_time},
 	{GUC_PREFIX ".repository_keepday", assign_int, &repository_keepday},
 	{GUC_PREFIX ".log_maintenance_command", assign_string, &log_maintenance_command},
+	{GUC_PREFIX ".controlfile_fsync_interval", assign_int, &controlfile_fsync_interval},
 	{":debug", assign_string, &msg_debug},
 	{":info", assign_string, &msg_info},
 	{":notice", assign_string, &msg_notice},
@@ -168,6 +170,7 @@ static struct ParamMap PARAM_MAP[] =
 	{":autoanalyze", assign_string, &msg_autoanalyze},
 	{":checkpoint_starting", assign_string, &msg_checkpoint_starting},
 	{":checkpoint_complete", assign_string, &msg_checkpoint_complete},
+	{":restartpoint_complete", assign_string, &msg_restartpoint_complete},
 	{NULL}
 };
 
@@ -184,7 +187,6 @@ isTTY(int fd)
 int
 main(int argc, char *argv[])
 {
-	PGconn	*conn;
 	void	*logger_retval;
 
 	shutdown_state = STARTUP;
@@ -243,29 +245,15 @@ main(int argc, char *argv[])
 	pgut_putenv("PGCONNECT_TIMEOUT", "2");
 	pgut_putenv("PGCLIENTENCODING", pg_encoding_to_char(server_encoding));
 
+	/* create lock file */
+	create_lock_file();
+
 	/*
 	 * set the abort level to FATAL so that the daemon should not be
 	 * terminated by ERRORs.
 	 */
 	pgut_abort_level = FATAL;
 
-	/* wait until repository database is ready to accept connection */
-	for (;;)
-	{
-		if (!postmaster_is_alive())
-			return STATSINFO_EXIT_FAILED;
-
-		if ((conn = pgut_connect(repository_server, NO, ERROR)))
-		{
-			/* check the state of repository database */
-			if (!check_repository(conn))
-				return STATSINFO_EXIT_FAILED;
-			break;
-		}
-		sleep(10);	/* 10s */
-	}
-	pgut_disconnect(conn);
-
 	/* init logger, collector, and writer module */
 	pthread_mutex_init(&shutdown_state_lock, NULL);
 	collector_init();
@@ -416,7 +404,7 @@ do_connect(PGconn **conn, const char *info, const char *schema)
 		return *conn;
 
 	pgut_disconnect(*conn);
-	*conn = pgut_connect(info, NO, DEBUG2);
+	*conn = pgut_connect(info, NO, ERROR);
 
 	if (PQstatus(*conn) == CONNECTION_OK)
 	{
@@ -438,7 +426,7 @@ do_connect(PGconn **conn, const char *info, const char *schema)
 /*
  * requires $PGDATA/contrib/pg_{schema}.sql
  */
-static bool
+bool
 ensure_schema(PGconn *conn, const char *schema)
 {
 	PGresult	   *res;
@@ -712,8 +700,6 @@ error:
 done:
 	termStringInfo(&name);
 	termStringInfo(&value);
-
-	after_readopt();
 }
 
 /*
@@ -734,32 +720,11 @@ readopt_from_db(PGresult *res)
 		assign_param(name, value);
 	}
 
-	after_readopt();
-
 	if (!(enable_maintenance & MAINTENANCE_MODE_SNAPSHOT))
 		elog(NOTICE,
 			"automatic maintenance is disable. Please note the data size of the repository");
 }
 
-/*
- * called after parameters are reloaded.
- */
-static void
-after_readopt(void)
-{
-	/*
-	 * We hope the message format for checkpoint starting ends with
-	 * "%s%s%s%s%s%s%s" on all locales.
-	 */
-	checkpoint_starting_prefix_len = 0;
-	if (msg_checkpoint_starting)
-	{
-		const char *flags = strstr(msg_checkpoint_starting, "%s%s%s%s%s%s%s");
-		if (flags)
-			checkpoint_starting_prefix_len = flags - msg_checkpoint_starting;
-	}
-}
-
 /*
  * delay unless shutdown is requested.
  */
@@ -936,51 +901,140 @@ execute_script(PGconn *conn, const char *script_file)
 }
 
 /*
- * check the state of repository database.
- *  - connect to the repository database
- *  - statsrepo schema version
+ * create the lock file.
+ * store our own PID in the lock file.
  */
-static bool
-check_repository(PGconn *conn)
+static void
+create_lock_file(void)
 {
-	PGresult	*res;
-	uint32		 version;
+	int		fd;
+	char	lockfile[MAXPGPATH];
+	char	buffer[64];
+	pid_t	my_pid;
 
-	/* check statsrepo schema exists */
-	res = pgut_execute(conn,
-		"SELECT nspname FROM pg_namespace WHERE nspname = 'statsrepo'", 0, NULL);
-	if (PQresultStatus(res) != PGRES_TUPLES_OK)
-		goto error;
-	if (PQntuples(res) == 0)
-		goto done;	/* not installed */
-	PQclear(res);
+	my_pid = getpid();
 
-	/* check the statsrepo schema version */
-	res = pgut_execute(conn, "SELECT statsrepo.get_version()", 0, NULL);
-	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	join_path_components(lockfile, data_directory, STATSINFO_LOCK_FILE);
+
+	/* create the lockfile */
+	fd = open(lockfile, O_WRONLY | O_CREAT | O_EXCL, 0600);
+	if (fd < 0)
 	{
-		char *sqlstate;
+		FILE	*fp;
+		pid_t	 lock_si_pid;
+		pid_t	 lock_pg_pid;
 
-		sqlstate = PQresultErrorField(res, PG_DIAG_SQLSTATE);
-		if (sqlstate && strcmp(sqlstate, "42883") == 0)	/* undefined function */
-			elog(ERROR, "incompatible statsrepo schema: version mismatch");
-		goto error;
+		if (errno != EEXIST && errno != EACCES)
+			ereport(FATAL,
+				(errcode_errno(),
+				 errmsg("could not create lock file \"%s\": %m", lockfile)));
+
+		/*
+		 * if lockfile already exists, do the check to avoid multiple boot.
+		 * and if another process still exists, terminate it.
+		 */
+		fd = open(lockfile, O_RDWR, 0600);
+		if (fd < 0)
+		{
+			ereport(FATAL,
+				(errcode_errno(),
+				 errmsg("could not open lock file \"%s\": %m", lockfile)));
+		}
+
+		fp = fdopen(fd, "r+");
+		if (fp == NULL)
+			ereport(FATAL,
+				(errcode_errno(),
+				 errmsg("could not open lock file \"%s\": %m", lockfile)));
+
+		errno = 0;
+		if (fscanf(fp, "%d\n%d\n", &lock_si_pid, &lock_pg_pid) != 2)
+		{
+			if (errno == 0)
+				elog(FATAL, "bogus data in lock file \"%s\"", lockfile);
+			else
+				ereport(FATAL,
+					(errcode_errno(),
+					 errmsg("could not read lock file \"%s\": %m", lockfile)));
+		}
+
+		if (kill(lock_si_pid, 0) == 0)	/* process is alive */
+		{
+			/* check the postmaster PID */
+			if (lock_pg_pid == postmaster_pid)
+				elog(FATAL, "is another pg_statsinfod (PID %d) running",
+					lock_si_pid);
+
+			/* terminate the another process still exists */
+			elog(NOTICE, "terminate the another process still exists (PID %d)",
+				lock_si_pid);
+			if (kill(lock_si_pid, SIGKILL) != 0)
+				elog(ERROR, "could not send kill signal (PID %d): %m",
+					lock_si_pid);
+		}
+
+		/* reset the seek position in order to write the lock file */
+		lseek(fd, (off_t) 0, SEEK_SET);
 	}
 
-	/* verify the version of statsrepo schema */
-	parse_uint32(PQgetvalue(res, 0, 0), &version);
-	if (version != STATSREPO_SCHEMA_VERSION &&
-		((version / 100) != (STATSREPO_SCHEMA_VERSION / 100)))
+	/* write content to the lockfile */
+	snprintf(buffer, sizeof(buffer), "%d\n%d\n", my_pid, postmaster_pid);
+
+	errno = 0;
+	if (write(fd, buffer, strlen(buffer)) != strlen(buffer))
 	{
-		elog(ERROR, "incompatible statsrepo schema: version mismatch");
-		goto error;
+		int		save_errno = errno;
+
+		close(fd);
+		unlink(lockfile);
+		/* if write didn't set errno, assume problem is no disk space */
+		errno = save_errno ? save_errno : ENOSPC;
+		ereport(FATAL,
+			(errcode_errno(),
+			 errmsg("could not write lock file \"%s\": %m", lockfile)));
+	}
+	if (fsync(fd) != 0)
+	{
+		int		save_errno = errno;
+
+		close(fd);
+		unlink(lockfile);
+		errno = save_errno;
+		ereport(FATAL,
+			(errcode_errno(),
+			 errmsg("could not write lock file \"%s\": %m", lockfile)));
+	}
+	if (close(fd) != 0)
+	{
+		int		save_errno = errno;
+
+		unlink(lockfile);
+		errno = save_errno;
+		ereport(FATAL,
+			(errcode_errno(),
+			 errmsg("could not write lock file \"%s\": %m", lockfile)));
 	}
 
-done:
-	PQclear(res);
-	return true;
+	/* automatic removal of the lock file at exit */
+	atexit(unlink_lock_file);
+}
 
-error:
-	PQclear(res);
-	return false;
+/*
+ * atexit() callback to remove a lock file.
+ */
+static void
+unlink_lock_file(void)
+{
+	char	lockfile[MAXPGPATH];
+
+	join_path_components(lockfile, data_directory, STATSINFO_LOCK_FILE);
+
+	if (unlink(lockfile) != 0)
+	{
+		if (errno != ENOENT)
+			ereport(FATAL,
+				(errcode_errno(),
+				 errmsg("could not remove lock file \"%s\": %m",
+				 	lockfile)));
+	}
 }
