@@ -11,6 +11,8 @@
 #define PG_STATSINFOD_H
 
 #include "postgres_fe.h"
+#include "miscadmin.h"
+#include "libpq/pqsignal.h"
 #include "pgut/pgut.h"
 #include "pgut/pgut-list.h"
 #include "pgut/pgut-pthread.h"
@@ -25,12 +27,12 @@
 #define SECS_PER_DAY		86400	/* seconds per day */
 
 #define STATSINFO_CONTROL_FILE		"pg_statsinfo.control"
-#define STATSINFO_CONTROL_VERSION	20400
+#define STATSINFO_CONTROL_VERSION	20500
 
 #define LOGGER_RETURN_SUCCESS		0x00
 #define LOGGER_RETURN_FAILED		0xff
 
-#define STATSREPO_SCHEMA_VERSION	20400
+#define STATSREPO_SCHEMA_VERSION	20500
 
 /* read settings */
 #define SQL_SELECT_CUSTOM_SETTINGS "\
@@ -67,10 +69,16 @@ FROM \
 		('" GUC_PREFIX ".enable_maintenance'), \
 		('" GUC_PREFIX ".maintenance_time'), \
 		('" GUC_PREFIX ".repository_keepday'), \
-		('" GUC_PREFIX ".log_maintenance_command')) AS t(name) \
+		('" GUC_PREFIX ".log_maintenance_command'), \
+		('" GUC_PREFIX ".controlfile_fsync_interval')) AS t(name) \
 	LEFT JOIN pg_settings s \
 	ON t.name = s.name"
 
+/* reworked from access/xlog_internal.h */
+#define XLogSegSize    ((uint32) XLOG_SEG_SIZE)
+#define XLogSegsPerFile (((uint32) 0xffffffff) / XLogSegSize)
+#define XLogFileSize  (XLogSegsPerFile * XLogSegSize)
+
 /* shutdown state */
 typedef enum ShutdownState
 {
@@ -82,6 +90,14 @@ typedef enum ShutdownState
 	LOGGER_SHUTDOWN
 } ShutdownState;
 
+/* writer state */
+typedef enum WriterState
+{
+	WRITER_READY,
+	WRITER_NORMAL,
+	WRITER_FALLBACK
+} WriterState;
+
 /* writer queue type */
 typedef enum WriterQueueType
 {
@@ -97,6 +113,7 @@ typedef enum WriterQueueType
 typedef enum StatsinfoState
 {
 	STATSINFO_RUNNING,
+	STATSINFO_STOPPED,
 	STATSINFO_SHUTDOWNED
 } StatsinfoState;
 
@@ -134,6 +151,7 @@ extern char		   *adjust_log_error;
 extern char		   *adjust_log_log;
 extern char		   *adjust_log_fatal;
 extern char		   *textlog_nologging_users;
+extern int			controlfile_fsync_interval;
 /*---- GUC variables (writer) ----------*/
 extern char		   *repository_server;
 extern int		    enable_maintenance;
@@ -158,10 +176,11 @@ extern char		   *msg_autovacuum;
 extern char		   *msg_autoanalyze;
 extern char		   *msg_checkpoint_starting;
 extern char		   *msg_checkpoint_complete;
-extern size_t		checkpoint_starting_prefix_len;
+extern char		   *msg_restartpoint_complete;
 /*--------------------------------------*/
 
 extern volatile ShutdownState	shutdown_state;
+extern volatile WriterState		writer_state;
 extern bool						shutdown_message_found;
 
 /* threads */
@@ -269,6 +288,7 @@ bool check_maintenance_log(pid_t log_maintenance_pid, int fd_err);
 /* pg_statsinfod.c */
 extern bool postmaster_is_alive(void);
 extern PGconn *do_connect(PGconn **conn, const char *info, const char *schema);
+extern bool ensure_schema(PGconn *conn, const char *schema);
 extern int str_to_elevel(const char *value);
 extern const char *elevel_to_str(int elevel);
 extern void shutdown_progress(ShutdownState state);
