@@ -17,11 +17,12 @@ CREATE SCHEMA statsrepo;
 
 CREATE TABLE statsrepo.instance
 (
-	instid			bigserial,
-	name			text NOT NULL,
-	hostname		text NOT NULL,
-	port			integer NOT NULL,
-	pg_version		text,
+	instid				bigserial,
+	name				text NOT NULL,
+	hostname			text NOT NULL,
+	port				integer NOT NULL,
+	pg_version			text,
+	xlog_file_size		bigint,
 	PRIMARY KEY (instid),
 	UNIQUE (name, hostname, port)
 );
@@ -197,6 +198,7 @@ CREATE TABLE statsrepo.activity
 	idle_in_xact		float8,
 	waiting				float8,
 	running				float8,
+	max_backends		integer,
 	max_xact_client		inet,
 	max_xact_pid		integer,
 	max_xact_start		timestamptz,
@@ -211,6 +213,7 @@ CREATE TABLE statsrepo.setting
 	snapid			bigint,
 	name			text,
 	setting			text,
+	unit			text,
 	source			text,
 	PRIMARY KEY (snapid, name),
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
@@ -479,7 +482,7 @@ $$ LANGUAGE sql;
 
 -- get_version() - version of statsrepo schema
 CREATE FUNCTION statsrepo.get_version() RETURNS text AS
-'SELECT CAST(''20400'' AS TEXT)'
+'SELECT CAST(''20500'' AS TEXT)'
 LANGUAGE sql IMMUTABLE;
 
 -- tps() - transaction per seconds
@@ -510,12 +513,12 @@ $$
 LANGUAGE sql IMMUTABLE STRICT;
 
 -- xlog_location_diff() - compute the difference in bytes between two WAL locations
-CREATE FUNCTION statsrepo.xlog_location_diff(text, text)
+CREATE FUNCTION statsrepo.xlog_location_diff(text, text, bigint)
 RETURNS numeric AS
 $$
 	/* XLogFileSize * (xlogid1 - xlogid2) + xrecoff1 - xrecoff2 */
 	SELECT
-		(X'FF000000'::bigint * (t.xlogid1 - t.xlogid2)::numeric + t.xrecoff1 - t.xrecoff2)
+		($3 * (t.xlogid1 - t.xlogid2)::numeric + t.xrecoff1 - t.xrecoff2)
 	FROM
 	(
 		SELECT
@@ -1386,7 +1389,7 @@ $$
 			xe.time,
 			xe.location,
 			xe.xlogfile,
-			statsrepo.xlog_location_diff(xe.location, xs.location) AS write_size,
+			statsrepo.xlog_location_diff(xe.location, xs.location, xe.xlog_file_size) AS write_size,
 			xe.time - xs.time AS duration
 		FROM
 		 	(SELECT
@@ -1394,13 +1397,16 @@ $$
 		 		s.time,
 		 		x.location,
 		 		x.xlogfile,
-				(SELECT max(snapid) FROM statsrepo.snapshot WHERE snapid < s.snapid AND instid = s.instid) AS prev_snapid
+				(SELECT max(snapid) FROM statsrepo.snapshot WHERE snapid < s.snapid AND instid = s.instid) AS prev_snapid,
+				i.xlog_file_size
 		 	 FROM
 			 	statsrepo.xlog x,
-				statsrepo.snapshot s
+				statsrepo.snapshot s,
+				statsrepo.instance i
 			 WHERE
 				x.snapid BETWEEN $1 AND $2
 				AND x.snapid = s.snapid
+				AND s.instid = i.instid
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)) AS xe,
 		 	(SELECT
 		 		s.snapid,
@@ -1661,6 +1667,41 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'CPU Usage + Load Average'
+CREATE FUNCTION statsrepo.get_cpu_loadavg_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT "user"			numeric,
+	OUT system			numeric,
+	OUT idle			numeric,
+	OUT iowait			numeric,
+	OUT loadavg1		numeric,
+	OUT loadavg5		numeric,
+	OUT loadavg15		numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		c.user,
+		c.system,
+		c.idle,
+		c.iowait,
+		l.loadavg1::numeric(6,3),
+		l.loadavg5::numeric(6,3),
+		l.loadavg15::numeric(6,3)
+	FROM
+		statsrepo.get_cpu_usage_tendency($1, $2) c,
+		statsrepo.loadavg l,
+		statsrepo.snapshot s
+	WHERE
+		c.snapid = l.snapid
+		AND c.snapid = s.snapid
+	ORDER BY
+		c.snapid;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'IO Usage'
 CREATE FUNCTION statsrepo.get_io_usage(
 	IN snapid_begin			bigint,
@@ -1676,45 +1717,27 @@ CREATE FUNCTION statsrepo.get_io_usage(
 ) RETURNS SETOF record AS
 $$
 	SELECT
-		a.device_name,
-		a.device_tblspaces,
-		statsrepo.sub(a.drs + o.drs_add, b.drs) / 2 / 1024,
-		statsrepo.sub(a.dws + o.dws_add, b.dws) / 2 / 1024,
-		statsrepo.sub(a.drt + o.drt_add, b.drt),
-		statsrepo.sub(a.dwt + o.dwt_add, b.dwt),
-		round((o.diq + b.diq) / (o.cnt + 1), 3),
-		statsrepo.sub(a.dit + o.dit_add, b.dit)
+		DISTINCT ON (de.device_name, de.device_major, de.device_minor)
+		coalesce(de.device_name, 'unknown'),
+		de.device_tblspaces,
+		CASE WHEN de.device_readsector IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_readsector + dx.drs_add, ds.device_readsector) / 2 / 1024 END,
+		CASE WHEN de.device_writesector IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_writesector + dx.dws_add, ds.device_writesector) / 2 / 1024 END,
+		CASE WHEN de.device_readtime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_readtime + dx.drt_add, ds.device_readtime) END,
+		CASE WHEN de.device_writetime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_writetime + dx.dwt_add, ds.device_writetime) END,
+		CASE WHEN de.device_ioqueue IS NULL THEN NULL ELSE
+			round((dx.diq + ds.device_ioqueue) / (dx.cnt + 1), 3) END,
+		CASE WHEN de.device_iototaltime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_iototaltime + dx.dit_add, ds.device_iototaltime) END
 	FROM
+		statsrepo.device de,
+		statsrepo.device ds,
 		(SELECT
-			snapid,
-			device_name,
-			device_tblspaces,
-			device_readsector as drs,
-			device_readtime as drt,
-			device_writesector as dws,
-			device_writetime as dwt, 
-			device_ioqueue as diq,
-			device_iototaltime as dit
-		 FROM
-		 	statsrepo.device
-		 WHERE
-		 	snapid = $1) b,
-		(SELECT
-			snapid,
-			device_name,
-			device_tblspaces,
-			device_readsector as drs,
-			device_readtime as drt,
-			device_writesector as dws,
-			device_writetime as dwt, 
-			device_ioqueue as diq,
-			device_iototaltime as dit
-		 FROM
-		 	statsrepo.device
-		 WHERE
-		 	snapid = $2) a,
-		(SELECT
-			d.device_name,
+			d.device_major,
+			d.device_minor,
 			(sum(d.overflow_drs) * 4294967296)::bigint AS drs_add,
 			(sum(d.overflow_drt) * 4294967296)::bigint AS drt_add,
 			(sum(d.overflow_dws) * 4294967296)::bigint AS dws_add,
@@ -1723,19 +1746,20 @@ $$
 			sum(d.device_ioqueue) AS diq,
 			count(*) AS cnt
 		 FROM
-			statsrepo.device d
-			LEFT JOIN statsrepo.snapshot s ON s.snapid = d.snapid
+			statsrepo.device d,
+			statsrepo.snapshot s
 		 WHERE
-			s.snapid > $1 AND s.snapid <= $2
+		 	d.snapid = s.snapid
+			AND s.snapid > $1 AND s.snapid <= $2
 			AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
 		 GROUP BY
-		 	d.device_name) o,
-		statsrepo.snapshot s
+		 	d.device_major, d.device_minor) dx
 	WHERE
-		a.snapid = s.snapid
-		AND a.device_name = b.device_name
-		AND a.device_name = o.device_name
-		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2);
+		(de.device_major = dx.device_major AND de.device_minor = dx.device_minor)
+		AND ds.snapid = $1
+		AND de.snapid = $2
+	ORDER BY
+		de.device_name;
 $$
 LANGUAGE sql;
 
@@ -1786,6 +1810,7 @@ $$
 				d.snapid BETWEEN $1 AND $2
 				AND d.snapid = s.snapid
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				d.snapid, d.device_name, s.time, s.instid) AS de,
 			(SELECT
@@ -1805,6 +1830,7 @@ $$
 				d.snapid BETWEEN $1 AND $2
 				AND d.snapid = s.snapid
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				d.snapid, d.device_name, s.time, s.instid) AS ds
 		WHERE
@@ -1866,6 +1892,7 @@ $$
 				d.snapid BETWEEN $1 AND $2
 				AND d.snapid = s.snapid
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				d.snapid, d.device_name, s.time, s.instid) AS de,
 			(SELECT
@@ -1885,6 +1912,7 @@ $$
 				d.snapid BETWEEN $1 AND $2
 				AND d.snapid = s.snapid
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				d.snapid, d.device_name, s.time, s.instid) AS ds
 		WHERE
@@ -1910,9 +1938,9 @@ CREATE FUNCTION statsrepo.get_loadavg_tendency(
 $$
 	SELECT
 		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
-		l.loadavg1::numeric(1000, 3),
-		l.loadavg5::numeric(1000, 3),
-		l.loadavg15::numeric(1000, 3)
+		l.loadavg1::numeric(6,3),
+		l.loadavg5::numeric(6,3),
+		l.loadavg15::numeric(6,3)
 	FROM
 		statsrepo.loadavg l,
 		statsrepo.snapshot s
@@ -2274,9 +2302,9 @@ CREATE FUNCTION statsrepo.get_autovacuum_activity(
 	OUT nspname			text,
 	OUT relname			text,
 	OUT "count"			bigint,
-	OUT avg_index_scans	numeric,
 	OUT avg_tup_removed	numeric,
 	OUT avg_tup_remain	numeric,
+	OUT avg_index_scans	numeric,
 	OUT avg_duration	numeric,
 	OUT max_duration	numeric
 ) RETURNS SETOF record AS
@@ -2286,9 +2314,9 @@ $$
 		schema,
 		"table",
 		count(*),
-		round(avg(index_scans)::numeric,3),
 		round(avg(tup_removed)::numeric,3),
 		round(avg(tup_remain)::numeric,3),
+		round(avg(index_scans)::numeric,3),
 		round(avg(duration)::numeric,3),
 		round(max(duration)::numeric,3)
 	FROM
@@ -2342,6 +2370,41 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Autovacuum Activity'
+CREATE FUNCTION statsrepo.get_autoanalyze_stats(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT datname			text,
+	OUT nspname			text,
+	OUT relname			text,
+	OUT total_duration	numeric,
+	OUT avg_duration	numeric,
+	OUT max_duration	numeric,
+	OUT "count"			bigint
+) RETURNS SETOF record AS
+$$
+	SELECT
+		database,
+		schema,
+		"table",
+		round(sum(duration)::numeric,3),
+		round(avg(duration)::numeric,3),
+		round(max(duration)::numeric,3),
+		count(*)
+	FROM
+		statsrepo.autoanalyze a,
+		(SELECT min(time) AS time FROM statsrepo.snapshot WHERE snapid >= $1) b,
+		(SELECT max(time) AS time FROM statsrepo.snapshot WHERE snapid <= $2) e
+	WHERE
+		a.start BETWEEN b.time AND e.time
+		AND a.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	GROUP BY
+		database, schema, "table"
+	ORDER BY
+		4 DESC;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Query Activity (Functions)'
 CREATE FUNCTION statsrepo.get_query_activity_functions(
 	IN snapid_begin		bigint,
@@ -2507,28 +2570,74 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Replication Delays'
+CREATE FUNCTION statsrepo.get_replication_delays(
+	IN snapid_begin			bigint,
+	IN snapid_end			bigint,
+	OUT "timestamp"			text,
+	OUT client				text,
+	OUT flush_delay_size	numeric,
+	OUT replay_delay_size	numeric,
+	OUT sync_state			text
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		host(r.client_addr) || ':' || r.client_port AS client,
+		statsrepo.xlog_location_diff(
+			split_part(r.current_location, ' ', 1),
+			split_part(r.flush_location, ' ', 1),
+			i.xlog_file_size),
+		statsrepo.xlog_location_diff(
+			split_part(r.current_location, ' ', 1),
+			split_part(r.replay_location, ' ', 1),
+			i.xlog_file_size),
+		(SELECT sync_state FROM statsrepo.replication WHERE snapid = $2
+			AND client_addr = r.client_addr AND client_port = r.client_port)
+	FROM
+		statsrepo.replication r,
+		statsrepo.snapshot s,
+		statsrepo.instance i
+	WHERE
+		r.snapid = s.snapid
+		AND s.instid = i.instid
+		AND r.snapid BETWEEN $1 AND $2
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+		AND r.flush_location IS NOT NULL
+		AND r.replay_location IS NOT NULL
+	ORDER BY
+		s.snapid, client;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Setting Parameters'
 CREATE FUNCTION statsrepo.get_setting_parameters(
 	IN snapid_begin	bigint,
 	IN snapid_end	bigint,
 	OUT name		text,
 	OUT setting		text,
+	OUT unit		text,
 	OUT source		text
 ) RETURNS SETOF record AS
 $$
 	SELECT
-		so.name,
-		CASE WHEN sa.setting = so.setting THEN
-			so.setting
+		se.name,
+		CASE WHEN se.setting <> sb.setting THEN
+			coalesce(sb.setting, '(default)') || ' -> ' || coalesce(se.setting, '(default)')
 		ELSE
-			coalesce(sa.setting, '(default)') || ' -> ' || coalesce(so.setting, '(default)')
+			se.setting
 		END,
-		so.source
+		se.unit,
+		se.source
 	FROM
-		statsrepo.setting so LEFT JOIN statsrepo.setting sa
-			ON so.name = sa.name AND sa.snapid = $1
+		statsrepo.setting sb,
+		statsrepo.setting se
 	WHERE
-		so.snapid = (SELECT MIN(snapid) FROM statsrepo.setting WHERE snapid >= $2);
+		sb.name = se.name
+		AND sb.snapid = $1
+		AND se.snapid = $2
+	ORDER BY
+		se.name;
 $$
 LANGUAGE sql;
 
