@@ -3,7 +3,7 @@
  *
  * Create a repository schema for PostgreSQL 8.4 and later.
  *
- * Copyright (c) 2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2011-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 -- Adjust this setting to control where the objects get created.
@@ -17,11 +17,12 @@ CREATE SCHEMA statsrepo;
 
 CREATE TABLE statsrepo.instance
 (
-	instid			bigserial,
-	name			text NOT NULL,
-	hostname		text NOT NULL,
-	port			integer NOT NULL,
-	pg_version		text,
+	instid				bigserial,
+	name				text NOT NULL,
+	hostname			text NOT NULL,
+	port				integer NOT NULL,
+	pg_version			text,
+	xlog_file_size		bigint,
 	PRIMARY KEY (instid),
 	UNIQUE (name, hostname, port)
 );
@@ -73,6 +74,11 @@ CREATE TABLE statsrepo.database
 	confl_snapshot		bigint,
 	confl_bufferpin		bigint,
 	confl_deadlock		bigint,
+	temp_files			bigint,
+	temp_bytes			bigint,
+	deadlocks			bigint,
+	blk_read_time		double precision,
+	blk_write_time		double precision,
 	PRIMARY KEY (snapid, dbid),
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
@@ -190,6 +196,7 @@ CREATE TABLE statsrepo.activity
 	idle_in_xact		float8,
 	waiting				float8,
 	running				float8,
+	max_backends		integer,
 	max_xact_client		inet,
 	max_xact_pid		integer,
 	max_xact_start		timestamptz,
@@ -204,6 +211,7 @@ CREATE TABLE statsrepo.setting
 	snapid			bigint,
 	name			text,
 	setting			text,
+	unit			text,
 	source			text,
 	PRIMARY KEY (snapid, name),
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
@@ -240,12 +248,16 @@ CREATE TABLE statsrepo.statement
 	rows				bigint,
 	shared_blks_hit		bigint,
 	shared_blks_read	bigint,
+	shared_blks_dirtied	bigint,
 	shared_blks_written	bigint,
 	local_blks_hit		bigint,
 	local_blks_read		bigint,
+	local_blks_dirtied	bigint,
 	local_blks_written	bigint,
 	temp_blks_read		bigint,
 	temp_blks_written	bigint,
+	blk_read_time		double precision,
+	blk_write_time		double precision,
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE,
 	FOREIGN KEY (snapid, dbid) REFERENCES statsrepo.database (snapid, dbid)
 );
@@ -260,8 +272,8 @@ CREATE TABLE statsrepo.function
 	funcname		name,
 	argtypes		text,
 	calls			bigint,
-	total_time		bigint,
-	self_time		bigint,
+	total_time		double precision,
+	self_time		double precision,
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE,
 	FOREIGN KEY (snapid, dbid) REFERENCES statsrepo.database (snapid, dbid)
 );
@@ -279,6 +291,11 @@ CREATE TABLE statsrepo.autovacuum
 	page_remain		integer,
 	tup_removed		bigint,
 	tup_remain		bigint,
+	page_hit		integer,
+	page_miss		integer,
+	page_dirty		integer,
+	read_rate		double precision,
+	write_rate		double precision,
 	duration		real,
 	FOREIGN KEY (instid) REFERENCES statsrepo.instance (instid) ON DELETE CASCADE
 );
@@ -350,6 +367,28 @@ CREATE TABLE statsrepo.device
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
 
+CREATE TABLE statsrepo.loadavg
+(
+	snapid			bigint,
+	loadavg1		real,
+	loadavg5		real,
+	loadavg15		real,
+	PRIMARY KEY (snapid),
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
+CREATE TABLE statsrepo.memory
+(
+	snapid		bigint,
+	memfree		bigint,
+	buffers		bigint,
+	cached		bigint,
+	swap		bigint,
+	dirty		bigint,
+	PRIMARY KEY (snapid),
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
 CREATE TABLE statsrepo.profile
 (
 	snapid			bigint,
@@ -401,6 +440,14 @@ CREATE TABLE statsrepo.replication
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
 
+CREATE TABLE statsrepo.xlog
+(
+	snapid		bigint,
+	location	text,
+	xlogfile	text,
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
 -- del_snapshot(snapid) - delete the specified snapshot.
 CREATE FUNCTION statsrepo.del_snapshot(bigint) RETURNS void AS
 $$
@@ -425,6 +472,11 @@ LANGUAGE sql;
 -- utility function for reporter.
 ------------------------------------------------------------------------------
 
+-- get_version() - version of statsrepo schema
+CREATE FUNCTION statsrepo.get_version() RETURNS text AS
+'SELECT CAST(''20500'' AS TEXT)'
+LANGUAGE sql IMMUTABLE;
+
 -- tps() - transaction per seconds
 CREATE FUNCTION statsrepo.tps(numeric, interval) RETURNS numeric AS
 'SELECT (CASE WHEN extract(epoch FROM $2) > 0 THEN $1 / extract(epoch FROM $2) ELSE 0 END)::numeric(1000, 3)'
@@ -440,6 +492,76 @@ CREATE FUNCTION statsrepo.sub(anyelement, anyelement) RETURNS anyelement AS
 'SELECT coalesce($1, 0) - coalesce($2, 0)'
 LANGUAGE sql;
 
+-- convert_hex() - convert a hexadecimal string to a decimal number
+CREATE FUNCTION statsrepo.convert_hex(text)
+RETURNS bigint AS
+$$
+	SELECT
+		(sum((16::numeric ^ (length($1) - i)) *
+			position(upper(substring($1 from i for 1)) in '123456789ABCDEF')))::bigint
+	FROM
+		generate_series(1, length($1)) AS t(i);
+$$
+LANGUAGE sql IMMUTABLE STRICT;
+
+-- xlog_location_diff() - compute the difference in bytes between two WAL locations
+CREATE FUNCTION statsrepo.xlog_location_diff(text, text, bigint)
+RETURNS numeric AS
+$$
+	/* XLogFileSize * (xlogid1 - xlogid2) + xrecoff1 - xrecoff2 */
+	SELECT
+		($3 * (t.xlogid1 - t.xlogid2)::numeric + t.xrecoff1 - t.xrecoff2)
+	FROM
+	(
+		SELECT
+			statsrepo.convert_hex(lsn1[1]) AS xlogid1,
+			statsrepo.convert_hex(lsn1[2]) AS xrecoff1,
+			statsrepo.convert_hex(lsn2[1]) AS xlogid2,
+			statsrepo.convert_hex(lsn2[2]) AS xrecoff2
+		 FROM
+			regexp_matches($1, '^([0-F]{1,8})/([0-F]{1,8})$') AS lsn1,
+			regexp_matches($2, '^([0-F]{1,8})/([0-F]{1,8})$') AS lsn2
+	) t;
+$$
+LANGUAGE sql IMMUTABLE STRICT;
+
+-- pg_size_pretty() - formatting with size units
+CREATE FUNCTION statsrepo.pg_size_pretty(bigint)
+RETURNS text AS
+$$
+DECLARE
+	size	bigint := $1;
+	buf		text;
+	limit1	bigint := 10 * 1024;
+	limit2	bigint := limit1 * 2 - 1;
+BEGIN
+	IF size < limit1 THEN
+		buf := size || ' bytes';
+	ELSE
+		size := size >> 9;	/* keep one extra bit for rounding */
+		IF size < limit2 THEN
+			buf := (size + 1) / 2 || ' KiB';
+		ELSE
+			size := size >> 10;
+			IF size < limit2 THEN
+				buf := (size + 1) / 2 || ' MiB';
+			ELSE
+				size := size >> 10;
+				IF size < limit2 THEN
+					buf := (size + 1) / 2 || ' GiB';
+				ELSE
+					size := size >> 10;
+					buf := (size + 1) / 2 || ' TiB';
+				END IF;
+			END IF;
+		END IF;
+	END IF;
+
+	RETURN buf;
+END;
+$$
+LANGUAGE plpgsql IMMUTABLE STRICT;
+
 -- tables - pre-JOINed tables
 CREATE VIEW statsrepo.tables AS
   SELECT t.snapid,
@@ -809,7 +931,7 @@ $$
 		statsrepo.snapshot b,
 		statsrepo.snapshot e,
 		(SELECT
-			pg_size_pretty(sum(ed.size)::int8),
+			statsrepo.pg_size_pretty(sum(ed.size)::int8),
 			sum(ed.xact_commit) - sum(sd.xact_commit),
 			sum(ed.xact_rollback) - sum(sd.xact_rollback)
 		 FROM
@@ -838,7 +960,12 @@ CREATE FUNCTION statsrepo.get_dbstats(
 	OUT blks_hit_rate		numeric,
 	OUT blks_hit_tps		numeric,
 	OUT blks_read_tps		numeric,
-	OUT tup_fetch_tps		numeric
+	OUT tup_fetch_tps		numeric,
+	OUT temp_files			bigint,
+	OUT temp_bytes			bigint,
+	OUT deadlocks			bigint,
+	OUT blk_read_time		numeric,
+	OUT blk_write_time		numeric
 ) RETURNS SETOF record AS
 $$
 	SELECT
@@ -865,7 +992,12 @@ $$
 		statsrepo.tps(
 			statsrepo.sub(ed.tup_returned, sd.tup_returned) +
 			statsrepo.sub(ed.tup_fetched, sd.tup_fetched),
-			es.time - ss.time)
+			es.time - ss.time),
+		statsrepo.sub(ed.temp_files, sd.temp_files),
+		statsrepo.sub(ed.temp_bytes, sd.temp_bytes) / 1024 / 1024,
+		statsrepo.sub(ed.deadlocks, sd.deadlocks),
+		statsrepo.sub(ed.blk_read_time, sd.blk_read_time)::numeric(1000, 3),
+		statsrepo.sub(ed.blk_write_time, sd.blk_write_time)::numeric(1000, 3)
 	FROM
 		statsrepo.snapshot ss,
 		statsrepo.snapshot es,
@@ -1111,10 +1243,14 @@ CREATE FUNCTION statsrepo.get_proc_tendency(
 $$
 	SELECT
 		a.snapid,
-		idle::numeric(1000, 3), 
-		idle_in_xact::numeric(1000, 3),
-		waiting::numeric(1000, 3),
-		running::numeric(1000, 3)
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle_in_xact / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * waiting / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * running / (idle + idle_in_xact + waiting + running))::numeric(5,2) END
 	FROM
 		statsrepo.activity a,
 		statsrepo.snapshot s
@@ -1141,10 +1277,14 @@ CREATE FUNCTION statsrepo.get_proc_tendency_report(
 $$
 	SELECT
 		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
-		idle::numeric(1000, 3), 
-		idle_in_xact::numeric(1000, 3),
-		waiting::numeric(1000, 3),
-		running::numeric(1000, 3)
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle_in_xact / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * waiting / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * running / (idle + idle_in_xact + waiting + running))::numeric(5,2) END
 	FROM
 		statsrepo.activity a,
 		statsrepo.snapshot s
@@ -1158,6 +1298,67 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'WAL Statistics'
+CREATE FUNCTION statsrepo.get_xlog_tendency(
+	IN snapid_begin			bigint,
+	IN snapid_end			bigint,
+	OUT "timestamp"			text,
+	OUT location			text,
+	OUT xlogfile			text,
+	OUT write_size			numeric,
+	OUT write_size_per_sec	numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(time, 'YYYY-MM-DD HH24:MI'),
+		location,
+		xlogfile,
+		(write_size / 1024 / 1024)::numeric(1000, 3),
+		(statsrepo.tps(write_size, duration) / 1024 / 1024)::numeric(1000, 3)
+	FROM
+	(
+		SELECT
+			s.snapid,
+			s.time,
+			x.location,
+			x.xlogfile,
+			statsrepo.xlog_location_diff(
+				x.location, lag(x.location) OVER w, i.xlog_file_size) AS write_size,
+			s.time - lag(s.time) OVER w AS duration
+		 FROM
+			statsrepo.xlog x,
+			statsrepo.snapshot s,
+			statsrepo.instance i
+		 WHERE
+			x.snapid BETWEEN $1 AND $2
+			AND x.snapid = s.snapid
+			AND s.instid = i.instid
+			AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+		 WINDOW w AS (ORDER BY s.snapid)
+		 ORDER BY
+		 	s.snapid
+	) t
+	WHERE
+		snapid > $1;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'WAL Statistics'
+CREATE FUNCTION statsrepo.get_xlog_stats(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT write_total		numeric,
+	OUT write_speed		numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		sum(write_size)::numeric(1000, 3),
+		avg(write_size_per_sec)::numeric(1000, 3)
+	FROM
+		statsrepo.get_xlog_tendency($1, $2);
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'CPU Usage'
 CREATE FUNCTION statsrepo.get_cpu_usage(
 	IN snapid_begin	bigint,
@@ -1308,6 +1509,41 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'CPU Usage + Load Average'
+CREATE FUNCTION statsrepo.get_cpu_loadavg_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT "user"			numeric,
+	OUT system			numeric,
+	OUT idle			numeric,
+	OUT iowait			numeric,
+	OUT loadavg1		numeric,
+	OUT loadavg5		numeric,
+	OUT loadavg15		numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		c.user,
+		c.system,
+		c.idle,
+		c.iowait,
+		l.loadavg1::numeric(6,3),
+		l.loadavg5::numeric(6,3),
+		l.loadavg15::numeric(6,3)
+	FROM
+		statsrepo.get_cpu_usage_tendency($1, $2) c,
+		statsrepo.loadavg l,
+		statsrepo.snapshot s
+	WHERE
+		c.snapid = l.snapid
+		AND c.snapid = s.snapid
+	ORDER BY
+		c.snapid;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'IO Usage'
 CREATE FUNCTION statsrepo.get_io_usage(
 	IN snapid_begin			bigint,
@@ -1323,45 +1559,27 @@ CREATE FUNCTION statsrepo.get_io_usage(
 ) RETURNS SETOF record AS
 $$
 	SELECT
-		a.device_name,
-		a.device_tblspaces,
-		statsrepo.sub(a.drs + o.drs_add, b.drs) / 2 / 1024,
-		statsrepo.sub(a.dws + o.dws_add, b.dws) / 2 / 1024,
-		statsrepo.sub(a.drt + o.drt_add, b.drt),
-		statsrepo.sub(a.dwt + o.dwt_add, b.dwt),
-		round((o.diq + b.diq) / (o.cnt + 1), 3),
-		statsrepo.sub(a.dit + o.dit_add, b.dit)
+		DISTINCT ON (de.device_name, de.device_major, de.device_minor)
+		coalesce(de.device_name, 'unknown'),
+		de.device_tblspaces,
+		CASE WHEN de.device_readsector IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_readsector + dx.drs_add, ds.device_readsector) / 2 / 1024 END,
+		CASE WHEN de.device_writesector IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_writesector + dx.dws_add, ds.device_writesector) / 2 / 1024 END,
+		CASE WHEN de.device_readtime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_readtime + dx.drt_add, ds.device_readtime) END,
+		CASE WHEN de.device_writetime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_writetime + dx.dwt_add, ds.device_writetime) END,
+		CASE WHEN de.device_ioqueue IS NULL THEN NULL ELSE
+			round((dx.diq + ds.device_ioqueue) / (dx.cnt + 1), 3) END,
+		CASE WHEN de.device_iototaltime IS NULL THEN NULL ELSE
+			statsrepo.sub(de.device_iototaltime + dx.dit_add, ds.device_iototaltime) END
 	FROM
+		statsrepo.device de,
+		statsrepo.device ds,
 		(SELECT
-			snapid,
-			device_name,
-			device_tblspaces,
-			device_readsector as drs,
-			device_readtime as drt,
-			device_writesector as dws,
-			device_writetime as dwt, 
-			device_ioqueue as diq,
-			device_iototaltime as dit
-		 FROM
-		 	statsrepo.device
-		 WHERE
-		 	snapid = $1) b,
-		(SELECT
-			snapid,
-			device_name,
-			device_tblspaces,
-			device_readsector as drs,
-			device_readtime as drt,
-			device_writesector as dws,
-			device_writetime as dwt, 
-			device_ioqueue as diq,
-			device_iototaltime as dit
-		 FROM
-		 	statsrepo.device
-		 WHERE
-		 	snapid = $2) a,
-		(SELECT
-			d.device_name,
+			d.device_major,
+			d.device_minor,
 			(sum(d.overflow_drs) * 4294967296)::bigint AS drs_add,
 			(sum(d.overflow_drt) * 4294967296)::bigint AS drt_add,
 			(sum(d.overflow_dws) * 4294967296)::bigint AS dws_add,
@@ -1370,19 +1588,20 @@ $$
 			sum(d.device_ioqueue) AS diq,
 			count(*) AS cnt
 		 FROM
-			statsrepo.device d
-			LEFT JOIN statsrepo.snapshot s ON s.snapid = d.snapid
+			statsrepo.device d,
+			statsrepo.snapshot s
 		 WHERE
-			s.snapid > $1 AND s.snapid <= $2
+		 	d.snapid = s.snapid
+			AND s.snapid > $1 AND s.snapid <= $2
 			AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
 		 GROUP BY
-		 	d.device_name) o,
-		statsrepo.snapshot s
+		 	d.device_major, d.device_minor) dx
 	WHERE
-		a.snapid = s.snapid
-		AND a.device_name = b.device_name
-		AND a.device_name = o.device_name
-		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2);
+		(de.device_major = dx.device_major AND de.device_minor = dx.device_minor)
+		AND ds.snapid = $1
+		AND de.snapid = $2
+	ORDER BY
+		de.device_name;
 $$
 LANGUAGE sql;
 
@@ -1435,6 +1654,7 @@ $$
 				s.snapid = d.snapid
 				AND s.snapid BETWEEN $1 AND $2
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				s.snapid, s.time, d.device_name) AS d
 		WINDOW w AS (PARTITION BY device_name ORDER BY snapid)
@@ -1496,6 +1716,7 @@ $$
 			 	s.snapid = d.snapid
 				AND s.snapid BETWEEN $1 AND $2
 				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+				AND d.device_name IS NOT NULL
 			 GROUP BY
 				s.snapid, s.time, d.device_name) AS d
 		WINDOW w AS (PARTITION BY device_name ORDER BY snapid)
@@ -1507,6 +1728,64 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Load Average'
+CREATE FUNCTION statsrepo.get_loadavg_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT "1min"			numeric,
+	OUT "5min"			numeric,
+	OUT "15min"			numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		l.loadavg1::numeric(6,3),
+		l.loadavg5::numeric(6,3),
+		l.loadavg15::numeric(6,3)
+	FROM
+		statsrepo.loadavg l,
+		statsrepo.snapshot s
+	WHERE
+		s.snapid BETWEEN $1 AND $2
+		AND s.snapid = l.snapid
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	ORDER BY
+		s.snapid;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'Memory Usage'
+CREATE FUNCTION statsrepo.get_memory_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT memfree			numeric,
+	OUT buffers			numeric,
+	OUT cached			numeric,
+	OUT swap			numeric,
+	OUT dirty			numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		(m.memfree::float / 1024)::numeric(1000, 2),
+		(m.buffers::float / 1024)::numeric(1000, 2),
+		(m.cached::float / 1024)::numeric(1000, 2),
+		(m.swap::float / 1024)::numeric(1000, 2),
+		(m.dirty::float / 1024)::numeric(1000, 2)
+	FROM
+		statsrepo.memory m,
+		statsrepo.snapshot s
+	WHERE
+		s.snapid BETWEEN $1 AND $2
+		AND s.snapid = m.snapid
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	ORDER BY
+		s.snapid;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Disk Usage per Tablespace'
 CREATE FUNCTION statsrepo.get_disk_usage_tablespace(
 	IN snapid_begin		bigint,
@@ -1779,7 +2058,6 @@ $$
 		AND i.snapid = c.snapid
 		AND i.tbl = c.tbl
 		AND c.attnum = ANY (i.indkey)
-		AND i.isclustered
 		AND i.schema NOT IN ('pg_catalog', 'pg_toast', 'information_schema')
 		AND c.correlation < 1
 	ORDER BY
@@ -1826,9 +2104,9 @@ CREATE FUNCTION statsrepo.get_autovacuum_activity(
 	OUT nspname			text,
 	OUT relname			text,
 	OUT "count"			bigint,
-	OUT avg_index_scans	numeric,
 	OUT avg_tup_removed	numeric,
 	OUT avg_tup_remain	numeric,
+	OUT avg_index_scans	numeric,
 	OUT avg_duration	numeric,
 	OUT max_duration	numeric
 ) RETURNS SETOF record AS
@@ -1838,9 +2116,9 @@ $$
 		schema,
 		"table",
 		count(*),
-		round(avg(index_scans)::numeric,3),
 		round(avg(tup_removed)::numeric,3),
 		round(avg(tup_remain)::numeric,3),
+		round(avg(index_scans)::numeric,3),
 		round(avg(duration)::numeric,3),
 		round(max(duration)::numeric,3)
 	FROM
@@ -1857,6 +2135,78 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Autovacuum Activity'
+CREATE FUNCTION statsrepo.get_autovacuum_activity2(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT datname			text,
+	OUT nspname			text,
+	OUT relname			text,
+	OUT avg_page_hit	numeric,
+	OUT avg_page_miss	numeric,
+	OUT avg_page_dirty	numeric,
+	OUT avg_read_rate	numeric,
+	OUT avg_write_rate	numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		database,
+		schema,
+		"table",
+		round(avg(page_hit)::numeric,3),
+		round(avg(page_miss)::numeric,3),
+		round(avg(page_dirty)::numeric,3),
+		round(avg(read_rate)::numeric,3),
+		round(avg(write_rate)::numeric,3)
+	FROM
+		statsrepo.autovacuum v,
+		(SELECT min(time) AS time FROM statsrepo.snapshot WHERE snapid >= $1) b,
+		(SELECT max(time) AS time FROM statsrepo.snapshot WHERE snapid <= $2) e
+	WHERE
+		v.start BETWEEN b.time AND e.time
+		AND v.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	GROUP BY
+		database, schema, "table"
+	ORDER BY
+		4 DESC;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'Autovacuum Activity'
+CREATE FUNCTION statsrepo.get_autoanalyze_stats(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT datname			text,
+	OUT nspname			text,
+	OUT relname			text,
+	OUT total_duration	numeric,
+	OUT avg_duration	numeric,
+	OUT max_duration	numeric,
+	OUT "count"			bigint
+) RETURNS SETOF record AS
+$$
+	SELECT
+		database,
+		schema,
+		"table",
+		round(sum(duration)::numeric,3),
+		round(avg(duration)::numeric,3),
+		round(max(duration)::numeric,3),
+		count(*)
+	FROM
+		statsrepo.autoanalyze a,
+		(SELECT min(time) AS time FROM statsrepo.snapshot WHERE snapid >= $1) b,
+		(SELECT max(time) AS time FROM statsrepo.snapshot WHERE snapid <= $2) e
+	WHERE
+		a.start BETWEEN b.time AND e.time
+		AND a.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	GROUP BY
+		database, schema, "table"
+	ORDER BY
+		4 DESC;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Query Activity (Functions)'
 CREATE FUNCTION statsrepo.get_query_activity_functions(
 	IN snapid_begin		bigint,
@@ -1866,8 +2216,8 @@ CREATE FUNCTION statsrepo.get_query_activity_functions(
 	OUT nspname			name,
 	OUT proname			name,
 	OUT calls			bigint,
-	OUT total_time		bigint,
-	OUT self_time		bigint,
+	OUT total_time		numeric,
+	OUT self_time		numeric,
 	OUT time_per_call	numeric
 ) RETURNS SETOF record AS
 $$
@@ -1877,10 +2227,10 @@ $$
 		s.name,
 		fe.funcname,
 		statsrepo.sub(fe.calls, fb.calls),
-		statsrepo.sub(fe.total_time, fb.total_time),
-		statsrepo.sub(fe.self_time, fb.self_time),
+		statsrepo.sub(fe.total_time, fb.total_time)::numeric(1000, 3),
+		statsrepo.sub(fe.self_time, fb.self_time)::numeric(1000, 3),
 		statsrepo.div(
-			statsrepo.sub(fe.total_time, fb.total_time),
+			statsrepo.sub(fe.total_time, fb.total_time)::numeric,
 			statsrepo.sub(fe.calls, fb.calls))
 	FROM
 		statsrepo.function fe LEFT JOIN statsrepo.function fb
@@ -1910,7 +2260,9 @@ CREATE FUNCTION statsrepo.get_query_activity_statements(
 	OUT query			text,
 	OUT calls			bigint,
 	OUT total_time		numeric,
-	OUT time_per_call	numeric
+	OUT time_per_call	numeric,
+	OUT blk_read_time	numeric,
+	OUT blk_write_time	numeric
 ) RETURNS SETOF record AS
 $$
 	SELECT
@@ -1921,7 +2273,9 @@ $$
 		statsrepo.sub(se.total_time, sb.total_time)::numeric(1000, 3),
 		statsrepo.div(
 			statsrepo.sub(se.total_time, sb.total_time)::numeric,
-			statsrepo.sub(se.calls, sb.calls))
+			statsrepo.sub(se.calls, sb.calls)),
+		statsrepo.sub(se.blk_read_time, sb.blk_read_time)::numeric(1000, 3),
+		statsrepo.sub(se.blk_write_time, sb.blk_write_time)::numeric(1000, 3)
 	FROM
 		statsrepo.statement se LEFT JOIN statsrepo.statement sb
 			ON sb.snapid = $1 AND sb.dbid = se.dbid AND sb.query = se.query,
@@ -1939,7 +2293,7 @@ $$
 $$
 LANGUAGE sql;
 
--- generate information that corresponds to 'Lock Activity'
+-- generate information that corresponds to 'Lock Conflicts'
 CREATE FUNCTION statsrepo.get_lock_activity(
 	IN snapid_begin		bigint,
 	IN snapid_end		bigint,
@@ -1985,7 +2339,7 @@ CREATE FUNCTION statsrepo.get_replication_activity(
 	OUT client_addr			inet,
 	OUT client_hostname		text,
 	OUT client_port			integer,
-	OUT backend_start		timestamptz,
+	OUT backend_start		timestamp,
 	OUT state				text,
 	OUT current_location	text,
 	OUT sent_location		text,
@@ -2002,7 +2356,7 @@ $$
 		client_addr,
 		client_hostname,
 		client_port,
-		backend_start,
+		backend_start::timestamp(0),
 		state,
 		current_location,
 		sent_location,
@@ -2018,28 +2372,74 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Replication Delays'
+CREATE FUNCTION statsrepo.get_replication_delays(
+	IN snapid_begin			bigint,
+	IN snapid_end			bigint,
+	OUT "timestamp"			text,
+	OUT client				text,
+	OUT flush_delay_size	numeric,
+	OUT replay_delay_size	numeric,
+	OUT sync_state			text
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		host(r.client_addr) || ':' || r.client_port AS client,
+		statsrepo.xlog_location_diff(
+			split_part(r.current_location, ' ', 1),
+			split_part(r.flush_location, ' ', 1),
+			i.xlog_file_size),
+		statsrepo.xlog_location_diff(
+			split_part(r.current_location, ' ', 1),
+			split_part(r.replay_location, ' ', 1),
+			i.xlog_file_size),
+		(SELECT sync_state FROM statsrepo.replication WHERE snapid = $2
+			AND client_addr = r.client_addr AND client_port = r.client_port)
+	FROM
+		statsrepo.replication r,
+		statsrepo.snapshot s,
+		statsrepo.instance i
+	WHERE
+		r.snapid = s.snapid
+		AND s.instid = i.instid
+		AND r.snapid BETWEEN $1 AND $2
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+		AND r.flush_location IS NOT NULL
+		AND r.replay_location IS NOT NULL
+	ORDER BY
+		s.snapid, client;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Setting Parameters'
 CREATE FUNCTION statsrepo.get_setting_parameters(
 	IN snapid_begin	bigint,
 	IN snapid_end	bigint,
 	OUT name		text,
 	OUT setting		text,
+	OUT unit		text,
 	OUT source		text
 ) RETURNS SETOF record AS
 $$
 	SELECT
-		so.name,
-		CASE WHEN sa.setting = so.setting THEN
-			so.setting
+		se.name,
+		CASE WHEN se.setting <> sb.setting THEN
+			coalesce(sb.setting, '(default)') || ' -> ' || coalesce(se.setting, '(default)')
 		ELSE
-			coalesce(sa.setting, '(default)') || ' -> ' || coalesce(so.setting, '(default)')
+			se.setting
 		END,
-		so.source
+		se.unit,
+		se.source
 	FROM
-		statsrepo.setting so LEFT JOIN statsrepo.setting sa
-			ON so.name = sa.name AND sa.snapid = $1
+		statsrepo.setting sb,
+		statsrepo.setting se
 	WHERE
-		so.snapid = (SELECT MIN(snapid) FROM statsrepo.setting WHERE snapid >= $2);
+		sb.name = se.name
+		AND sb.snapid = $1
+		AND se.snapid = $2
+	ORDER BY
+		se.name;
 $$
 LANGUAGE sql;
 
