@@ -25,7 +25,7 @@ INSERT INTO statsrepo.tablespace VALUES \
 
 #define SQL_INSERT_ACTIVITY "\
 INSERT INTO statsrepo.activity VALUES \
-($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)"
+($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)"
 
 #define SQL_INSERT_STATEMENT "\
 INSERT INTO statsrepo.statement \
@@ -44,7 +44,7 @@ INSERT INTO statsrepo.replication VALUES \
 INSERT INTO statsrepo.xlog VALUES ($1, $2, $3)"
 
 #define SQL_INSERT_SETTING "\
-INSERT INTO statsrepo.setting VALUES ($1, $2, $3, $4)"
+INSERT INTO statsrepo.setting VALUES ($1, $2, $3, $4, $5)"
 
 #define SQL_INSERT_ROLE "\
 INSERT INTO statsrepo.role VALUES ($1, $2, $3)"
