@@ -46,13 +46,19 @@
 #include "utils/rel.h"
 #endif
 
+#if PG_VERSION_NUM >= 90300
+#include "access/htup_details.h"
+#endif
+
 #include "pgut/pgut-be.h"
 #include "pgut/pgut-spi.h"
 #include "../common.h"
 
-#ifndef WIN32
-#include "linux/version.h"
-#endif
+#define INVALID_PID			(-1)
+#define START_WAIT_MIN		(10)
+#define START_WAIT_MAX		(300)
+#define STOP_WAIT_MIN		(10)
+#define STOP_WAIT_MAX		(300)
 
 /* also adjust non-critial setting parameters? */
 /* #define ADJUST_NON_CRITICAL_SETTINGS */
@@ -89,7 +95,7 @@
 	"pages: %d removed, %d remain\n" \
 	"tuples: %.0f removed, %.0f remain\n" \
 	"buffer usage: %d hits, %d misses, %d dirtied\n" \
-	"avg read rate: %.3f MiB/s, avg write rate: %.3f MiB/s\n" \
+	"avg read rate: %.3f %s, avg write rate: %.3f %s\n" \
 	"system usage: %s"
 #else
 #define MSG_AUTOVACUUM \
@@ -103,11 +109,11 @@
 #define MSG_AUTOANALYZE \
 	"automatic analyze of table \"%s.%s.%s\" system usage: %s"
 
-/* log_checkpoints: staring */
+/* log_checkpoint: staring */
 #define MSG_CHECKPOINT_STARTING \
-	"checkpoint starting:%s%s%s%s%s%s%s"
+	"%s starting: %s"
 
-/* log_checkpoints: complete */
+/* log_checkpoint: complete */
 #if PG_VERSION_NUM >= 90100
 #define MSG_CHECKPOINT_COMPLETE \
 	"checkpoint complete: wrote %d buffers (%.1f%%); " \
@@ -121,6 +127,19 @@
 	"write=%ld.%03d s, sync=%ld.%03d s, total=%ld.%03d s"
 #endif
 
+/* log_restartpoint: complete */
+#if PG_VERSION_NUM >= 90100
+#define MSG_RESTARTPOINT_COMPLETE \
+	"restartpoint complete: wrote %d buffers (%.1f%%); " \
+	"%d transaction log file(s) added, %d removed, %d recycled; " \
+	"write=%ld.%03d s, sync=%ld.%03d s, total=%ld.%03d s; " \
+	"sync files=%d, longest=%ld.%03d s, average=%ld.%03d s"
+#else
+#define MSG_RESTARTPOINT_COMPLETE \
+	"restartpoint complete: wrote %d buffers (%.1f%%); " \
+	"write=%ld.%03d s, sync=%ld.%03d s, total=%ld.%03d s"
+#endif
+
 PG_MODULE_MAGIC;
 
 static const char *
@@ -140,21 +159,22 @@ default_log_maintenance_command(void)
 
 /*---- GUC variables ----*/
 
-#define DEFAULT_SAMPLING_INTERVAL		5		/* sec */
-#define DEFAULT_SNAPSHOT_INTERVAL		600		/* sec */
-#define DEFAULT_SYSLOG_LEVEL			DISABLE
-#define DEFAULT_TEXTLOG_LEVEL			WARNING
-#define DEFAULT_MAINTENANCE_TIME		"00:02:00"
-#define DEFAULT_REPOSITORY_KEEPDAY		7		/* day */
-#define DEFAULT_LOG_MAINTENANCE_COMMAND	default_log_maintenance_command()
-#define DEFAULT_LONG_LOCK_THREASHOLD	30		/* sec */
-#define DEFAULT_STAT_STATEMENTS_MAX		30
-#define LONG_TRANSACTION_THRESHOLD		1.0		/* sec */
+#define DEFAULT_SAMPLING_INTERVAL			5		/* sec */
+#define DEFAULT_SNAPSHOT_INTERVAL			600		/* sec */
+#define DEFAULT_SYSLOG_LEVEL				DISABLE
+#define DEFAULT_TEXTLOG_LEVEL				WARNING
+#define DEFAULT_MAINTENANCE_TIME			"00:02:00"
+#define DEFAULT_REPOSITORY_KEEPDAY			7		/* day */
+#define DEFAULT_LOG_MAINTENANCE_COMMAND		default_log_maintenance_command()
+#define DEFAULT_LONG_LOCK_THREASHOLD		30		/* sec */
+#define DEFAULT_STAT_STATEMENTS_MAX			30
+#define DEFAULT_CONTROLFILE_FSYNC_INTERVAL	60		/* sec */
+#define LONG_TRANSACTION_THRESHOLD			1.0		/* sec */
 
 #if PG_VERSION_NUM < 80400
-#define DEFAULT_ENABLE_MAINTENANCE		"3"		/* snapshot + log */
+#define DEFAULT_ENABLE_MAINTENANCE			"3"		/* snapshot + log */
 #else
-#define DEFAULT_ENABLE_MAINTENANCE		"on"	/* snapshot + log */
+#define DEFAULT_ENABLE_MAINTENANCE			"on"	/* snapshot + log */
 #endif
 
 static const struct config_enum_entry elevel_options[] =
@@ -217,6 +237,7 @@ static char	   *log_maintenance_command = NULL;
 static int		long_lock_threashold = DEFAULT_LONG_LOCK_THREASHOLD;
 static int		stat_statements_max = DEFAULT_STAT_STATEMENTS_MAX;
 static char	   *stat_statements_exclude_users = NULL;
+static int		controlfile_fsync_interval = DEFAULT_CONTROLFILE_FSYNC_INTERVAL;
 
 /*---- Function declarations ----*/
 
@@ -225,6 +246,8 @@ PG_FUNCTION_INFO_V1(statsinfo_activity);
 PG_FUNCTION_INFO_V1(statsinfo_snapshot);
 PG_FUNCTION_INFO_V1(statsinfo_maintenance);
 PG_FUNCTION_INFO_V1(statsinfo_tablespaces);
+PG_FUNCTION_INFO_V1(statsinfo_start);
+PG_FUNCTION_INFO_V1(statsinfo_stop);
 PG_FUNCTION_INFO_V1(statsinfo_restart);
 PG_FUNCTION_INFO_V1(statsinfo_cpustats);
 PG_FUNCTION_INFO_V1(statsinfo_cpustats_noarg);
@@ -239,6 +262,8 @@ extern Datum PGUT_EXPORT statsinfo_activity(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_snapshot(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_maintenance(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_tablespaces(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT statsinfo_start(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT statsinfo_stop(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_restart(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_cpustats(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_cpustats_noarg(PG_FUNCTION_ARGS);
@@ -258,8 +283,9 @@ extern PGUT_EXPORT void	fini_last_xact_activity(void);
 static void inet_to_cstring(const SockAddr *addr, char host[NI_MAXHOST]);
 static void StartStatsinfoLauncher(void);
 static void StatsinfoLauncherMain(void);
+static void sil_sigusr1_handler(SIGNAL_ARGS);
+static void sil_sigusr2_handler(SIGNAL_ARGS);
 static void sil_sigchld_handler(SIGNAL_ARGS);
-static void sil_exit(SIGNAL_ARGS);
 static pid_t exec_background_process(char cmd[]);
 static uint64 get_sysident(void);
 static void must_be_superuser(void);
@@ -271,17 +297,8 @@ static pid_t get_postmaster_pid(void);
 static bool verify_log_filename(const char *filename);
 static bool verify_timestr(const char *timestr);
 static bool postmaster_is_alive(void);
-
-#if PG_VERSION_NUM >= 90100
-static bool check_textlog_filename(char **newval, void **extra, GucSource source);
-static bool check_enable_maintenance(char **newval, void **extra, GucSource source);
-static bool check_maintenance_time(char **newval, void **extra, GucSource source);
-#else
-static const char *assign_textlog_filename(const char *newval, bool doit, GucSource source);
-static const char *assign_enable_maintenance(const char *newval, bool doit, GucSource source);
-static const char *assign_maintenance_time(const char *newval, bool doit, GucSource source);
-#endif
-
+static bool is_shared_preload(const char *library);
+static pid_t get_statsinfo_pid(const char *pid_file);
 static Datum get_cpustats(FunctionCallInfo fcinfo,
 	int64 prev_cpu_user, int64 prev_cpu_system, int64 prev_cpu_idle, int64 prev_cpu_iowait);
 static Datum get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats);
@@ -289,23 +306,12 @@ static int exec_grep(const char *filename, const char *regex, List **records);
 static int exec_split(const char *rawstring, const char *regex, List **fields);
 static bool parse_int64(const char *value, int64 *result);
 static bool parse_float8(const char *value, double *result);
+static char *b_trim(char *str);
+static HeapTupleHeader search_devicestats(ArrayType *devicestats, const char *device_name);
 
 #if PG_VERSION_NUM < 80400
 static bool parse_bool(const char *value, bool *result);
 static const char *elevel_to_str(int elevel);
-#endif
-
-static char *b_trim(char *str);
-static Datum BuildArrayType(List *values, Oid elmtype, Datum(*convert)(void *));
-static Datum _CStringGetTextDatum(void *ptr);
-static HeapTupleHeader search_devicestats(ArrayType *devicestats, const char *device_name);
-
-#if PG_VERSION_NUM < 80400 || defined(WIN32)
-static int str_to_elevel(const char *name, const char *str,
-						 const struct config_enum_entry *options);
-#endif
-
-#if PG_VERSION_NUM < 80400
 static char	   *syslog_min_messages_str;
 static char	   *textlog_min_messages_str;
 static const char *assign_syslog_min_messages(const char *newval, bool doit, GucSource source);
@@ -315,7 +321,21 @@ static const char *assign_elevel(const char *name, int *var, const char *newval,
 /* 8.3 or earlier versions can work only with PGC_USERSET */
 #undef PGC_SIGHUP
 #define PGC_SIGHUP		PGC_USERSET
+#endif
 
+#if PG_VERSION_NUM < 80400 || defined(WIN32)
+static int str_to_elevel(const char *name, const char *str,
+						 const struct config_enum_entry *options);
+#endif
+
+#if PG_VERSION_NUM >= 90100
+static bool check_textlog_filename(char **newval, void **extra, GucSource source);
+static bool check_enable_maintenance(char **newval, void **extra, GucSource source);
+static bool check_maintenance_time(char **newval, void **extra, GucSource source);
+#else
+static const char *assign_textlog_filename(const char *newval, bool doit, GucSource source);
+static const char *assign_enable_maintenance(const char *newval, bool doit, GucSource source);
+static const char *assign_maintenance_time(const char *newval, bool doit, GucSource source);
 #endif
 
 /* sampled statistics */
@@ -328,6 +348,7 @@ typedef struct Stats
 	int			idle_in_xact;
 	int			waiting;
 	int			running;
+	int			max_backends;
 
 	/* longest transaction */
 	int			max_xact_pid;
@@ -339,9 +360,11 @@ typedef struct Stats
 
 static Stats	*stats;
 
-/* flags for pg_statsinfo launcher */
-static volatile bool need_exit = false;
-static volatile bool got_SIGCHLD = false;
+/* variables for pg_statsinfo launcher */
+static volatile bool	got_SIGCHLD = false;
+static volatile bool	got_SIGUSR1 = false;
+static volatile bool	got_SIGUSR2 = false;
+static pid_t			sil_pid = INVALID_PID;
 
 /*
  * statsinfo_sample - sample statistics for server instance.
@@ -457,6 +480,9 @@ statsinfo_sample(PG_FUNCTION_ARGS)
 	stats->waiting += waiting;
 	stats->running += running;
 
+	if (stats->max_backends < (backends - 1))
+		stats->max_backends = backends - 1;
+
 	stats->samples++;
 
 	PG_RETURN_VOID();
@@ -489,7 +515,7 @@ inet_to_cstring(const SockAddr *addr, char host[NI_MAXHOST])
 	}
 }
 
-#define NUM_ACTIVITY_COLS		9
+#define NUM_ACTIVITY_COLS		10
 
 /*
  * statsinfo_activity - accumulate sampled counters.
@@ -522,6 +548,7 @@ statsinfo_activity(PG_FUNCTION_ARGS)
 		values[i++] = Float8GetDatum(stats->idle_in_xact / samples);
 		values[i++] = Float8GetDatum(stats->waiting / samples);
 		values[i++] = Float8GetDatum(stats->running / samples);
+		values[i++] = Int32GetDatum(stats->max_backends);
 
 		if (stats->max_xact_client[0])
 			values[i++] = CStringGetTextDatum(stats->max_xact_client);
@@ -660,7 +687,7 @@ _PG_init(void)
 							   "Sets the latest file name for textlog.",
 							   NULL,
 							   &textlog_filename,
-							   "postgresql.log",
+							   "pg_statsinfo.log",
 							   PGC_SIGHUP,
 							   GUC_SUPERUSER_ONLY,
 #if PG_VERSION_NUM >= 90100
@@ -986,6 +1013,21 @@ _PG_init(void)
 							   NULL,
 							   NULL);
 
+	DefineCustomIntVariable(GUC_PREFIX ".controlfile_fsync_interval",
+							"sets the fsync interval of the control file.",
+							NULL,
+							&controlfile_fsync_interval,
+							DEFAULT_CONTROLFILE_FSYNC_INTERVAL,
+							-1,
+							INT_MAX,
+							PGC_SIGHUP,
+							GUC_UNIT_S,
+#if PG_VERSION_NUM >= 90100
+							NULL,
+#endif
+							NULL,
+							NULL);
+
 	EmitWarningsOnPlaceholders("pg_statsinfo");
 
 	if (IsUnderPostmaster)
@@ -1060,6 +1102,157 @@ _PG_fini(void)
 #endif
 }
 
+/*
+ * statsinfo_start - start statsinfo background process.
+ */
+Datum
+statsinfo_start(PG_FUNCTION_ARGS)
+{
+	int32	timeout;
+	char	pid_file[MAXPGPATH];
+	pid_t	pid;
+	int		cnt;
+	int		save_client_min_messages = client_min_messages;
+	int		save_log_min_messages = log_min_messages;
+
+	/*
+	 * adjust elevel to LOG so that message be output to client console only,
+	 * not write to server log.
+	 */
+	client_min_messages = LOG;
+	log_min_messages = FATAL;
+
+	must_be_superuser();
+
+	if (PG_ARGISNULL(0))
+		elog(ERROR, "argument must not be NULL");
+
+	timeout = PG_GETARG_INT32(0);
+	if (timeout < START_WAIT_MIN || timeout > START_WAIT_MAX)
+		elog(ERROR, "%d is outside the valid range for parameter (%d .. %d)",
+			timeout, START_WAIT_MIN, START_WAIT_MAX);
+
+	/* pg_statsinfo library must been preload as shared library */
+	if (!is_shared_preload("pg_statsinfo"))
+		elog(ERROR, "pg_statsinfo is not preloaded as shared library");
+
+	Assert(sil_pid != INVALID_PID);
+
+	join_path_components(pid_file, DataDir, STATSINFO_LOCK_FILE);
+
+	if ((pid = get_statsinfo_pid(pid_file)) != 0)	/* pid file exists */
+	{
+		if (kill(pid, 0) == 0)	/* process is alive */
+		{
+			elog(WARNING, "pg_statsinfod (PID %d) might be running", pid);
+			goto done;
+		}
+
+		/* remove PID file */
+		if (unlink(pid_file) != 0)
+			elog(ERROR, "could not remove file \"%s\": %s",
+				pid_file, strerror(errno));
+	}
+
+	/* send signal that instruct start the statsinfo background process */
+	if (kill(sil_pid, SIGUSR2) != 0)
+		elog(ERROR, "could not send start signal (PID %d): %m", sil_pid);
+
+	elog(LOG, "waiting for pg_statsinfod to start");
+
+	pid = get_statsinfo_pid(pid_file);
+	for (cnt = 0; pid == 0 && cnt < timeout; cnt++)
+	{
+		pg_usleep(1000000);		/* 1 sec */
+		pid = get_statsinfo_pid(pid_file);
+	}
+
+	if (pid == 0)	/* pid file still not exists */
+		elog(WARNING, "timed out waiting for pg_statsinfod startup");
+	else
+		elog(LOG, "pg_statsinfod started");
+
+done:
+	client_min_messages = save_client_min_messages;
+	log_min_messages = save_log_min_messages;
+	PG_RETURN_VOID();
+}
+
+/*
+ * statsinfo_stop - stop statsinfo background process.
+ */
+Datum
+statsinfo_stop(PG_FUNCTION_ARGS)
+{
+	int32	timeout;
+	char	pid_file[MAXPGPATH];
+	pid_t	pid;
+	int		cnt;
+	int		save_client_min_messages = client_min_messages;
+	int		save_log_min_messages = log_min_messages;
+
+	/*
+	 * adjust elevel to LOG so that message be output to client console only,
+	 * not write to server log.
+	 */
+	client_min_messages = LOG;
+	log_min_messages = FATAL;
+
+	must_be_superuser();
+
+	if (PG_ARGISNULL(0))
+		elog(ERROR, "argument must not be NULL");
+
+	timeout = PG_GETARG_INT32(0);
+	if (timeout < STOP_WAIT_MIN || timeout > STOP_WAIT_MAX)
+		elog(ERROR, "%d is outside the valid range for parameter (%d .. %d)",
+			timeout, STOP_WAIT_MIN, STOP_WAIT_MAX);
+
+	/* pg_statsinfo library must been preload as shared library */
+	if (!is_shared_preload("pg_statsinfo"))
+		elog(ERROR, "pg_statsinfo is not preloaded as shared library");
+
+	Assert(sil_pid != INVALID_PID);
+
+	join_path_components(pid_file, DataDir, STATSINFO_LOCK_FILE);
+
+	if ((pid = get_statsinfo_pid(pid_file)) == 0)	/* no pid file */
+	{
+		elog(WARNING, "PID file \"%s\" does not exist; is pg_statsinfod running?",
+			pid_file);
+		goto done;
+	}
+
+	if (kill(pid, 0) != 0)	/* process is not alive */
+	{
+		elog(WARNING, "pg_statsinfod is not running (PID %d)", pid);
+		goto done;
+	}
+
+	/* send signal that instruct stop the statsinfo background process */
+	if (kill(sil_pid, SIGUSR1) != 0)
+		elog(ERROR, "could not send stop signal (PID %d): %m", sil_pid);
+
+	elog(LOG, "waiting for pg_statsinfod to shut down");
+
+	pid = get_statsinfo_pid(pid_file);
+	for (cnt = 0; pid != 0 && cnt < timeout; cnt++)
+	{
+		pg_usleep(1000000);		/* 1 sec */
+		pid = get_statsinfo_pid(pid_file);
+	}
+
+	if (pid != 0)	/* pid file still exists */
+		elog(WARNING, "timed out waiting for pg_statsinfod shut down");
+	else
+		elog(LOG, "pg_statsinfod stopped");
+
+done:
+	client_min_messages = save_client_min_messages;
+	log_min_messages = save_log_min_messages;
+	PG_RETURN_VOID();
+}
+
 /*
  * statsinfo_restart - Restart statsinfo background process.
  */
@@ -1114,7 +1307,7 @@ statsinfo_restart(PG_FUNCTION_ARGS)
 #define NUM_STAT_FIELDS_MIN		6
 
 /* not support a kernel that does not have the required fields at "/proc/stat" */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,41)
+#if !LINUX_VERSION_AT_LEAST(2,5,41)
 #error kernel version 2.5.41 or later is required
 #endif
 
@@ -1271,12 +1464,14 @@ statsinfo_devicestats_noarg(PG_FUNCTION_ARGS)
 #define NUM_DISKSTATS_PARTITION_FIELDS	7
 #define SQL_SELECT_TABLESPACES "\
 SELECT \
-	device, name \
+	device, \
+	statsinfo.array_agg(name) \
 FROM \
 	statsinfo.tablespaces \
 WHERE \
 	device IS NOT NULL \
-ORDER BY device"
+GROUP BY \
+	device"
 
 #define ARRNELEMS(x)	ArrayGetNItems(ARR_NDIM(x), ARR_DIMS(x))
 #define ARRPTR(x)		((HeapTupleHeader) ARR_DATA_PTR(x))
@@ -1295,9 +1490,8 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 	SPITupleTable	*tuptable;
 	Datum			 values[NUM_DEVICESTATS_COLS];
 	bool			 nulls[NUM_DEVICESTATS_COLS];
-	List			*spclist = NIL;
-	char			*prev_device = NULL;
 	int				 row;
+	bool			 isnull;
 
 	/* check to see if caller supports us returning a tuplestore */
 	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
@@ -1333,9 +1527,10 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 
 	for (row = 0; row < SPI_processed; row++)
 	{
+		HeapTuple tup = tuptable->vals[row];
+		TupleDesc desc = tuptable->tupdesc;
 		HeapTupleHeader prev_devicestats;
 		char *device;
-		char *spcname;
 		char *dev_major;
 		char *dev_minor;
 		char *dev_name = NULL;
@@ -1352,20 +1547,7 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 		List *fields = NIL;
 		int   nfield;
 
-		device = SPI_getvalue(tuptable->vals[row], tuptable->tupdesc, 1);
-		spcname = SPI_getvalue(tuptable->vals[row], tuptable->tupdesc, 2);
-
-		if (prev_device)
-		{
-			if (strcmp(device, prev_device) == 0)
-			{
-				spclist = lappend(spclist, spcname);
-				continue;
-			}
-			/* device_tblspaces */
-			values[14] = BuildArrayType(spclist, TYPE_DEVICE_TABLESPACES, _CStringGetTextDatum);
-			tuplestore_putvalues(tupstore, tupdesc, values, nulls);
-		}
+		device = SPI_getvalue(tup, desc, 1);
 
 		/* <device_mejor>:<device_minor> */
 		exec_split(device, ":", &devicenum);
@@ -1373,6 +1555,12 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 		dev_major = (char *) list_nth(devicenum, 0);
 		dev_minor = (char *) list_nth(devicenum, 1);
 
+		memset(nulls, 0, sizeof(nulls));
+		memset(values, 0, sizeof(values));
+		values[0] = CStringGetTextDatum(dev_major);			/* device_major */
+		values[1] = CStringGetTextDatum(dev_minor);			/* device_minor */
+		values[14] = SPI_getbinval(tup, desc, 2, &isnull);	/* device_tblspaces */
+
 		snprintf(regex, lengthof(regex), "^\\s*%s\\s+%s\\s+", dev_major, dev_minor);
 
 		/* extract device information */
@@ -1380,9 +1568,22 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 		{
 			ereport(DEBUG2,
 				(errmsg("device information of \"%s\" used by tablespace \"%s\" does not exist in \"%s\"",
-					device, spcname, FILE_DISKSTATS)));
-			prev_device = NULL;
-			spclist = list_truncate(spclist, 0);
+					device, SPI_getvalue(tup, desc, 2), FILE_DISKSTATS)));
+
+			nulls[2] = true;	/* device_name */
+			nulls[3] = true;	/* device_readsector */
+			nulls[4] = true;	/* device_readtime */
+			nulls[5] = true;	/* device_writesector */
+			nulls[6] = true;	/* device_writetime */
+			nulls[7] = true;	/* device_queue */
+			nulls[8] = true;	/* device_iototaltime */
+			nulls[9] = true;	/* overflow_drs */
+			nulls[10] = true;	/* overflow_drt */
+			nulls[11] = true;	/* overflow_dws */
+			nulls[12] = true;	/* overflow_dwt */
+			nulls[13] = true;	/* overflow_dit */
+
+			tuplestore_putvalues(tupstore, tupdesc, values, nulls);
 			continue;
 		}
 
@@ -1390,77 +1591,37 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 
 		nfield = exec_split(record, "\\s+", &fields);
 
-		memset(nulls, 0, sizeof(nulls));
-		memset(values, 0, sizeof(values));
-		spclist = list_truncate(spclist, 0);
-
 		if (nfield  == NUM_DISKSTATS_FIELDS)
 		{
-			/* device_major */
-			values[0] = CStringGetTextDatum(dev_major);
-
-			/* device_minor */
-			values[1] = CStringGetTextDatum(dev_minor);
-
-			/* device_name */
 			dev_name = list_nth(fields, 2);
-			values[2] = CStringGetTextDatum(dev_name);
-
-			/* device_readsector */
 			parse_int64(list_nth(fields, 5), &readsector);
-			values[3] = Int64GetDatum(readsector);
-
-			/* device_readtime */
 			parse_int64(list_nth(fields, 6), &readtime);
-			values[4] = Int64GetDatum(readtime);
-
-			/* device_writesector */
 			parse_int64(list_nth(fields, 9), &writesector);
-			values[5] = Int64GetDatum(writesector);
-
-			/* device_writetime */
 			parse_int64(list_nth(fields, 10), &writetime);
-			values[6] = Int64GetDatum(writetime);
-
-			/* device_queue */
 			parse_int64(list_nth(fields, 11), &ioqueue);
-			values[7] = Int64GetDatum(ioqueue);
-
-			/* device_iototaltime */
 			parse_int64(list_nth(fields, 13), &iototaltime);
-			values[8] = Int64GetDatum(iototaltime);
+
+			values[2] = CStringGetTextDatum(dev_name);	/* device_name */
+			values[3] = Int64GetDatum(readsector);		/* device_readsector */
+			values[4] = Int64GetDatum(readtime);		/* device_readtime */
+			values[5] = Int64GetDatum(writesector);		/* device_writesector */
+			values[6] = Int64GetDatum(writetime);		/* device_writetime */
+			values[7] = Int64GetDatum(ioqueue);			/* device_queue */
+			values[8] = Int64GetDatum(iototaltime);		/* device_iototaltime */
 		}
 		else if (nfield == NUM_DISKSTATS_PARTITION_FIELDS)
 		{
-			/* device_major */
-			values[0] = CStringGetTextDatum(dev_major);
-
-			/* device_minor */
-			values[1] = CStringGetTextDatum(dev_minor);
-
-			/* device_name */
 			dev_name = list_nth(fields, 2);
-			values[2] = CStringGetTextDatum(dev_name);
-
-			/* device_readsector */
 			parse_int64(list_nth(fields, 4), &readsector);
-			values[3] = Int64GetDatum(readsector);
-
-			/* device_readtime */
-			nulls[4] = true;
-
-			/* device_writesector */
 			parse_int64(list_nth(fields, 6), &writesector);
-			values[5] = Int64GetDatum(writesector);
 
-			/* device_writetime */
-			nulls[6] = true;
-
-			/* device_queue */
-			nulls[7] = true;
-
-			/* device_iototaltime */
-			nulls[8] = true;
+			values[2] = CStringGetTextDatum(dev_name);	/* device_name */
+			values[3] = Int64GetDatum(readsector);		/* device_readsector */
+			nulls[4] = true;							/* device_readtime */
+			values[5] = Int64GetDatum(writesector);		/* device_writesector */
+			nulls[6] = true;							/* device_writetime */
+			nulls[7] = true;							/* device_queue */
+			nulls[8] = true;							/* device_iototaltime */
 		}
 		else
 			ereport(ERROR,
@@ -1478,7 +1639,6 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 			int64 prev_writesector;
 			int64 prev_writetime;
 			int64 prev_iototaltime;
-			bool isnull;
 
 			prev_readsector = DatumGetInt64(GetAttributeByNum(prev_devicestats, 2, &isnull));
 			prev_readtime = DatumGetInt64(GetAttributeByNum(prev_devicestats, 3, &isnull));
@@ -1489,60 +1649,20 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 			/* overflow_drs */
 			if (readsector < prev_readsector)
 				values[9] = Int16GetDatum(1);
-			else
-				values[9] = Int16GetDatum(0);
-
 			/* overflow_drt */
 			if (nfield  == NUM_DISKSTATS_FIELDS && readtime < prev_readtime)
 				values[10] = Int16GetDatum(1);
-			else
-				values[10] = Int16GetDatum(0);
-
 			/* overflow_dws */
 			if (writesector < prev_writesector)
 				values[11] = Int16GetDatum(1);
-			else
-				values[11] = Int16GetDatum(0);
-
 			/* overflow_dwt */
 			if (nfield  == NUM_DISKSTATS_FIELDS && writetime < prev_writetime)
 				values[12] = Int16GetDatum(1);
-			else
-				values[12] = Int16GetDatum(0);
-
 			/* overflow_dit */
 			if (nfield  == NUM_DISKSTATS_FIELDS && iototaltime < prev_iototaltime)
 				values[13] = Int16GetDatum(1);
-			else
-				values[13] = Int16GetDatum(0);
-		}
-		else
-		{
-			/* overflow_drs */
-			values[9] = Int16GetDatum(0);
-
-			/* overflow_drt */
-			values[10] = Int16GetDatum(0);
-
-			/* overflow_dws */
-			values[11] = Int16GetDatum(0);
-
-			/* overflow_dwt */
-			values[12] = Int16GetDatum(0);
-
-			/* overflow_dit */
-			values[13] = Int16GetDatum(0);
 		}
 
-		spclist = lappend(spclist, spcname);
-		prev_device = device;
-	}
-
-	/* store the last tuple */
-	if (list_length(spclist) > 0)
-	{
-		/* device_tblspaces */
-		values[14] = BuildArrayType(spclist, TYPE_DEVICE_TABLESPACES, _CStringGetTextDatum);
 		tuplestore_putvalues(tupstore, tupdesc, values, nulls);
 	}
 
@@ -1929,7 +2049,7 @@ StartStatsinfoLauncher(void)
 	/*
 	 * invoke pg_statsinfo launcher processs
 	 */
-	switch (fork_process())
+	switch ((sil_pid = fork_process()))
 	{
 		case -1:
 			ereport(LOG,
@@ -1976,14 +2096,8 @@ StatsinfoLauncherMain(void)
 		(errmsg("pg_statsinfo launcher started")));
 
 	/* Set up signal handlers */
-	pqsignal(SIGHUP, sil_exit);
-	pqsignal(SIGINT, sil_exit);
-	pqsignal(SIGTERM, sil_exit);
-	pqsignal(SIGQUIT, sil_exit);
-	pqsignal(SIGALRM, sil_exit);
-	pqsignal(SIGPIPE, sil_exit);
-	pqsignal(SIGUSR1, sil_exit);
-	pqsignal(SIGUSR2, sil_exit);
+	pqsignal(SIGUSR1, sil_sigusr1_handler);
+	pqsignal(SIGUSR2, sil_sigusr2_handler);
 	pqsignal(SIGCHLD, sil_sigchld_handler);
 
 	/* launch a pg_statsinfod process */
@@ -1996,49 +2110,82 @@ StatsinfoLauncherMain(void)
 		if (!postmaster_is_alive())
 			break;
 
-		/* have received a signal that terminate process */
-		if (need_exit)
-			break;
+		/* instruct to stop pg_statsinfod process */
+		if (got_SIGUSR1)
+		{
+			got_SIGUSR1 = false;
+
+			if (StatsinfoPID == INVALID_PID)	/* not running */
+			{
+				ereport(WARNING,
+					(errmsg("pg_statsinfod is not running")));
+				continue;
+			}
+
+			/* send signal that instruct to shut down */
+			if (kill(StatsinfoPID, SIGUSR1) != 0)
+				ereport(WARNING,
+					(errmsg("could not send stop signal (PID %d): %m",
+						StatsinfoPID)));
+		}
+
+		/* instruct to start pg_statsinfod process */
+		if (got_SIGUSR2)
+		{
+			got_SIGUSR2 = false;
+
+			if (StatsinfoPID != INVALID_PID)	/* already running */
+			{
+				ereport(WARNING,
+					(errmsg("another pg_statsinfod might be running")));
+				continue;
+			}
+
+			/* launch the pg_statsinfo background process */
+			StatsinfoPID = exec_background_process(cmd);
+			launch_time = (pg_time_t) time(NULL);
+			launch_retry = 0;
+		}
 
 		/* pg_statsinfod process died */
 		if (got_SIGCHLD)
 		{
 			int status;
 
+			got_SIGCHLD = false;
+
 			waitpid(StatsinfoPID, &status, WNOHANG);
+			StatsinfoPID = INVALID_PID;
 
-			/* pg_statsinfod normally end, terminate the pg_statsinfo launcher */
-			if (status == 0)
-				break;
-
-			/* 
-			 * if the pg_statsinfod was aborted with fatal error,
-			 * then terminate the pg_statsinfo launcher
-			 */
-			if (WIFEXITED(status) && WEXITSTATUS(status) == STATSINFO_EXIT_FAILED)
+			if (WIFEXITED(status))
 			{
-				ereport(WARNING,
-					(errmsg("pg_statsinfod is aborted with fatal error, "
-							"terminate the pg_statsinfo launcher")));
-				break;
+				/* pg_statsinfod is normally end */
+				if (WEXITSTATUS(status) == STATSINFO_EXIT_SUCCESS)
+					continue;
+
+				/* pg_statsinfod is aborted with fatal error */
+				if (WEXITSTATUS(status) == STATSINFO_EXIT_FAILED)
+				{
+					ereport(WARNING,
+						(errmsg("pg_statsinfod is aborted with fatal error")));
+					continue;
+				}
 			}
 
-			/* pg_statsinfod abnormally end, relaunch new pg_statsinfod process */
-			ereport(WARNING,
-				(errmsg("pg_statsinfod is aborted")));
-
 			/* 
+			 * pg_statsinfod abnormally end, relaunch new pg_statsinfod process.
 			 * if the pg_statsinfod was aborted continuously,
-			 * then terminate the pg_statsinfo launcher
+			 * then not relaunch pg_statsinfod process.
 			 */
+			ereport(WARNING, (errmsg("pg_statsinfod is aborted")));
+
 			if (((pg_time_t) time(NULL) - launch_time) <= LAUNCH_RETRY_PERIOD)
 			{
 				if (launch_retry >= LAUNCH_RETRY_MAX)
 				{
 					ereport(WARNING,
-					(errmsg("pg_statsinfod is aborted continuously, "
-							"terminate the pg_statsinfo launcher")));
-					break;
+						(errmsg("pg_statsinfod is aborted continuously")));
+					continue;
 				}
 			}
 			else
@@ -2047,10 +2194,8 @@ StatsinfoLauncherMain(void)
 			ereport(LOG,
 				(errmsg("relaunch a pg_statsinfod process")));
 
-			got_SIGCHLD = false;
 			StatsinfoPID = exec_background_process(cmd);
 			launch_time = (pg_time_t) time(NULL);
-
 			launch_retry++;
 		}
 
@@ -2127,12 +2272,27 @@ exec_background_process(char cmd[])
 	send_str(fd, ":autoanalyze", _(MSG_AUTOANALYZE));
 	send_str(fd, ":checkpoint_starting", _(MSG_CHECKPOINT_STARTING));
 	send_str(fd, ":checkpoint_complete", _(MSG_CHECKPOINT_COMPLETE));
+	send_str(fd, ":restartpoint_complete", _(MSG_RESTARTPOINT_COMPLETE));
 	send_end(fd);
 	close(fd);
 
 	return fpid;
 }
 
+/* SIGUSR1: instructs to stop the pg_statsinfod process */
+static void
+sil_sigusr1_handler(SIGNAL_ARGS)
+{
+	got_SIGUSR1 = true;
+}
+
+/* SIGUSR2: instructs to start the pg_statsinfod process */
+static void
+sil_sigusr2_handler(SIGNAL_ARGS)
+{
+	got_SIGUSR2 = true;
+}
+
 /* SIGCHLD: pg_statsinfod process died */
 static void
 sil_sigchld_handler(SIGNAL_ARGS)
@@ -2140,12 +2300,6 @@ sil_sigchld_handler(SIGNAL_ARGS)
 	got_SIGCHLD = true;
 }
 
-static void
-sil_exit(SIGNAL_ARGS)
-{
-	need_exit = true;
-}
-
 /*
  * Read control file. We cannot retrieve it from "Control File" shared memory
  * because the shared memory might not be initialized yet.
@@ -3164,37 +3318,6 @@ b_trim(char *str)
 	return str;
 }
 
-static Datum
-BuildArrayType(List *values, Oid elmtype, Datum(*convert)(void *))
-{
-	ArrayType	*array_t;
-	ListCell	*cell;
-	Datum		*elems;
-	int16		 typlen;
-	bool		 typbyval;
-	char		 typalign;
-	int			 val_size;
-	int			 i;
-
-	val_size = list_length(values);
-	get_typlenbyvalalign(elmtype, &typlen, &typbyval, &typalign);
-
-	elems = palloc(sizeof(Datum) * val_size + 1);
-
-	i = 0;
-	foreach(cell, values)
-		elems[i++] = convert(lfirst(cell));
-
-	array_t = construct_array(elems, val_size, elmtype, typlen, typbyval, typalign);
-	return PointerGetDatum(array_t);
-}
-
-static Datum
-_CStringGetTextDatum(void *ptr)
-{
-	return CStringGetTextDatum((char *) ptr);
-}
-
 static HeapTupleHeader
 search_devicestats(ArrayType *devicestats, const char *device_name)
 {
@@ -3263,3 +3386,59 @@ postmaster_is_alive(void)
 	return (WaitForSingleObject(PostmasterHandle, 0) == WAIT_TIMEOUT);
 #endif   /* WIN32 */
 }
+
+/*
+ * is_shared_preload - check the library is preloaded as shared library
+ */
+static bool
+is_shared_preload(const char *library)
+{
+	char		*rawstring;
+	List		*elemlist;
+	ListCell	*cell;
+	bool		 find = false;
+
+	if (shared_preload_libraries_string == NULL ||
+		shared_preload_libraries_string[0] == '\0')
+		return false;
+
+	/* need a modifiable copy of string */
+	rawstring = pstrdup(shared_preload_libraries_string);
+
+	/* parse string into list of identifiers */
+	SplitIdentifierString(rawstring, ',', &elemlist);
+
+	foreach (cell, elemlist)
+	{
+		if (strcmp((char *) lfirst(cell), library) == 0)
+		{
+			find = true;
+			break;
+		}
+	}
+
+	pfree(rawstring);
+	list_free(elemlist);
+	return find;
+}
+
+static pid_t
+get_statsinfo_pid(const char *pid_file)
+{
+	FILE	*fp;
+	pid_t	 pid;
+
+	if ((fp = fopen(pid_file, "r")) == NULL)
+	{
+		/* No pid file, not an error */
+		if (errno == ENOENT)
+			return 0;
+		elog(ERROR,
+			"could not open PID file \"%s\": %s", pid_file, strerror(errno));
+	}
+
+	if (fscanf(fp, "%d\n", &pid) != 1)
+		elog(ERROR, "invalid data in PID file \"%s\"", pid_file);
+	fclose(fp);
+	return pid;
+}
