@@ -18,13 +18,33 @@
 #define SQL_SELECT_XACT_TENDENCY				"SELECT * FROM statsrepo.get_xact_tendency_report($1, $2)"
 #define SQL_SELECT_DBSIZE_TENDENCY				"SELECT * FROM statsrepo.get_dbsize_tendency_report($1, $2)"
 #define SQL_SELECT_RECOVERY_CONFLICTS 			"SELECT * FROM statsrepo.get_recovery_conflicts($1, $2)"
-#define SQL_SELECT_INSTANCE_PROC_RATIO			"SELECT * FROM statsrepo.get_proc_ratio($1, $2)"
-#define SQL_SELECT_INSTANCE_PROC_TENDENCY		"SELECT * FROM statsrepo.get_proc_tendency_report($1, $2)"
+#define SQL_SELECT_INSTANCE_PROC_TENDENCY "\
+SELECT * FROM statsrepo.get_proc_tendency_report($1, $2) \
+UNION ALL \
+SELECT \
+	'Average', \
+	avg(idle)::numeric(5,2), \
+	avg(idle_in_xact)::numeric(5,2), \
+	avg(waiting)::numeric(5,2), \
+	avg(running)::numeric(5,2) \
+FROM \
+	statsrepo.get_proc_tendency_report($1, $2)"
 #define SQL_SELECT_WALSTATS						"SELECT * FROM statsrepo.get_xlog_stats($1, $2)"
 #define SQL_SELECT_WALSTATS_TENDENCY			"SELECT * FROM statsrepo.get_xlog_tendency($1, $2)"
-#define SQL_SELECT_CPU_USAGE					"SELECT * FROM statsrepo.get_cpu_usage($1, $2)"
-#define SQL_SELECT_CPU_USAGE_TENDENCY			"SELECT * FROM statsrepo.get_cpu_usage_tendency_report($1, $2)"
-#define SQL_SELECT_LOADAVG_TENDENCY				"SELECT * FROM statsrepo.get_loadavg_tendency($1, $2)"
+#define SQL_SELECT_CPU_LOADAVG_TENDENCY "\
+SELECT * FROM statsrepo.get_cpu_loadavg_tendency($1, $2) \
+UNION ALL \
+SELECT \
+	'Average', \
+	avg(\"user\")::numeric(5,2), \
+	avg(system)::numeric(5,2), \
+	avg(idle)::numeric(5,2), \
+	avg(iowait)::numeric(5,2), \
+	avg(loadavg1)::numeric(6,3), \
+	avg(loadavg5)::numeric(6,3), \
+	avg(loadavg15)::numeric(6,3) \
+FROM \
+	statsrepo.get_cpu_loadavg_tendency($1, $2)"
 #define SQL_SELECT_MEMORY_TENDENCY				"SELECT * FROM statsrepo.get_memory_tendency($1, $2)"
 #define SQL_SELECT_IO_USAGE						"SELECT * FROM statsrepo.get_io_usage($1, $2)"
 #define SQL_SELECT_IO_USAGE_TENDENCY			"SELECT * FROM statsrepo.get_io_usage_tendency_report($1, $2)"
@@ -40,9 +60,9 @@
 SELECT \
 	datname || '.' || nspname || '.' || relname, \
 	\"count\", \
-	avg_index_scans, \
 	avg_tup_removed, \
 	avg_tup_remain, \
+	avg_index_scans, \
 	avg_duration, \
 	max_duration \
 FROM \
@@ -57,10 +77,60 @@ SELECT \
 	avg_write_rate \
 FROM \
 	statsrepo.get_autovacuum_activity2($1, $2)"
+#define SQL_SELECT_AUTOANALYZE_STATS "\
+SELECT \
+	datname || '.' || nspname || '.' || relname, \
+	total_duration, \
+	avg_duration, \
+	max_duration, \
+	\"count\" \
+FROM \
+	statsrepo.get_autoanalyze_stats($1, $2)"
 #define SQL_SELECT_QUERY_ACTIVITY_FUNCTIONS		"SELECT * FROM statsrepo.get_query_activity_functions($1, $2) LIMIT 20"
 #define SQL_SELECT_QUERY_ACTIVITY_STATEMENTS	"SELECT * FROM statsrepo.get_query_activity_statements($1, $2) LIMIT 20"
 #define SQL_SELECT_LOCK_CONFLICTS				"SELECT * FROM statsrepo.get_lock_activity($1, $2) LIMIT 20"
-#define SQL_SELECT_REPLICATION_ACTIVITY			"SELECT * FROM statsrepo.get_replication_activity($1, $2)"
+#define SQL_SELECT_REPLICATION_STATUS			"SELECT * FROM statsrepo.get_replication_activity($1, $2)"
+#define SQL_SELECT_REPLICATION_DELAYS_SYNC "\
+SELECT \
+	timestamp, \
+	client, \
+	flush_delay_size, \
+	replay_delay_size \
+FROM \
+	statsrepo.get_replication_delays($1, $2) \
+WHERE \
+	sync_state = 'sync'"
+#define SQL_SELECT_REPLICATION_DELAYS_ASYNC "\
+SELECT \
+	timestamp, \
+	client, \
+	flush_delay_size, \
+	replay_delay_size \
+FROM \
+	statsrepo.get_replication_delays($1, $2) \
+WHERE \
+	client = \
+	( \
+		SELECT \
+			host(r.client_addr) || ':' || r.client_port \
+		FROM \
+			statsrepo.replication r, \
+			statsrepo.snapshot s, \
+			statsrepo.instance i \
+		WHERE \
+			r.snapid = s.snapid \
+			AND s.instid = i.instid \
+			AND r.snapid = $2 \
+			AND r.sync_state != 'sync' \
+			AND r.flush_location IS NOT NULL \
+			AND r.replay_location IS NOT NULL \
+		ORDER BY \
+			statsrepo.xlog_location_diff( \
+				split_part(r.current_location, ' ', 1), \
+				split_part(r.flush_location, ' ', 1), \
+				i.xlog_file_size) DESC, client_addr, client_port \
+		LIMIT 1 \
+	)"
 #define SQL_SELECT_SETTING_PARAMETERS			"SELECT * FROM statsrepo.get_setting_parameters($1, $2)"
 #define SQL_SELECT_SCHEMA_INFORMATION_TABLES	"SELECT * FROM statsrepo.get_schema_info_tables($1, $2)"
 #define SQL_SELECT_SCHEMA_INFORMATION_INDEXES	"SELECT * FROM statsrepo.get_schema_info_indexes($1, $2)"
@@ -431,18 +501,6 @@ report_instance_activity(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "\n");
 	PQclear(res);
 
-	fprintf(out, "/** Instance Processes Ratio **/\n");
-	fprintf(out, "-----------------------------------\n");
-
-	res = pgut_execute(conn, SQL_SELECT_INSTANCE_PROC_RATIO, lengthof(params), params);
-	if (PQntuples(res) == 0)
-		return;
-	fprintf(out, "Back-end Idle Ratio          : %s %%\n", PQgetvalue(res, 0, 0));
-	fprintf(out, "Back-end Idle In Xact Ratio  : %s %%\n", PQgetvalue(res, 0, 1));
-	fprintf(out, "Back-end Waiting Ratio       : %s %%\n", PQgetvalue(res, 0, 2));
-	fprintf(out, "Back-end Running Ratio       : %s %%\n\n", PQgetvalue(res, 0, 3));
-	PQclear(res);
-
 	fprintf(out, "/** Instance Processes **/\n");
 	fprintf(out, "-----------------------------------\n");
 	fprintf(out, "%-16s  %12s  %12s  %12s  %12s\n",
@@ -477,50 +535,24 @@ report_resource_usage(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "/* OS Resource Usage */\n");
 	fprintf(out, "----------------------------------------\n\n");
 
-	fprintf(out, "/** CPU Usage **/\n");
+	fprintf(out, "/** CPU Usage + Load Average **/\n");
 	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-16s  %8s  %8s  %8s  %8s  %8s  %8s  %9s\n",
+		"DateTime", "User", "System", "Idle", "IOwait", "Loadavg1", "Loadavg5", "Loadavg15");
+	fprintf(out, "------------------------------------------------------------------------------------------\n");
 
-	res = pgut_execute(conn, SQL_SELECT_CPU_USAGE, lengthof(params), params);
-	if (PQntuples(res) == 0)
-		return;
-	fprintf(out, "User Mode Ratio    : %s %%\n", PQgetvalue(res, 0, 0));
-	fprintf(out, "System Mode Ratio  : %s %%\n", PQgetvalue(res, 0, 1));
-	fprintf(out, "Idle Mode Ratio    : %s %%\n", PQgetvalue(res, 0, 2));
-	fprintf(out, "IOwait Mode Ratio  : %s %%\n\n", PQgetvalue(res, 0, 3));
-	PQclear(res);
-
-	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %10s  %10s  %10s  %10s\n",
-		"DateTime", "User", "System", "Idle", "IOwait");
-	fprintf(out, "-------------------------------------------------------------------\n");
-
-	res = pgut_execute(conn, SQL_SELECT_CPU_USAGE_TENDENCY, lengthof(params), params);
+	res = pgut_execute(conn, SQL_SELECT_CPU_LOADAVG_TENDENCY, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %8s %%  %8s %%  %8s %%  %8s %%\n",
+		fprintf(out, "%-16s  %6s %%  %6s %%  %6s %%  %6s %%  %8s  %8s  %9s\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
 			PQgetvalue(res, i, 3),
-			PQgetvalue(res, i, 4));
-	}
-	fprintf(out, "\n");
-	PQclear(res);
-
-	fprintf(out, "/** Load Average **/\n");
-	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %7s  %7s  %7s\n",
-		"DateTime", "1 Min", "5 Min", "15 Min");
-	fprintf(out, "----------------------------------------------\n");
-
-	res = pgut_execute(conn, SQL_SELECT_LOADAVG_TENDENCY, lengthof(params), params);
-	for(i = 0; i < PQntuples(res); i++)
-	{
-		fprintf(out, "%-16s  %7s  %7s  %7s\n",
-			PQgetvalue(res, i, 0),
-			PQgetvalue(res, i, 1),
-			PQgetvalue(res, i, 2),
-			PQgetvalue(res, i, 3));
+			PQgetvalue(res, i, 4),
+			PQgetvalue(res, i, 5),
+			PQgetvalue(res, i, 6),
+			PQgetvalue(res, i, 7));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
@@ -535,15 +567,26 @@ report_resource_usage(PGconn *conn, ReportScope *scope, FILE *out)
 	res = pgut_execute(conn, SQL_SELECT_IO_USAGE, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-12s  %-24s  %8s MiB  %8s MiB  %14s ms  %14s ms  %16s  %12s ms\n",
-			PQgetvalue(res, i, 0),
-			PQgetvalue(res, i, 1),
-			PQgetvalue(res, i, 2),
-			PQgetvalue(res, i, 3),
-			PQgetvalue(res, i, 4),
-			PQgetvalue(res, i, 5),
-			PQgetvalue(res, i, 6),
-			PQgetvalue(res, i, 7));
+		if (strcmp(PQgetvalue(res, i, 0), "unknown") == 0)
+			fprintf(out, "%-12s  %-24s  %12s  %12s  %17s  %17s  %16s  %15s\n",
+				PQgetvalue(res, i, 0),
+				PQgetvalue(res, i, 1),
+				"(N/A)",
+				"(N/A)",
+				"(N/A)",
+				"(N/A)",
+				"(N/A)",
+				"(N/A)");
+		else
+			fprintf(out, "%-12s  %-24s  %8s MiB  %8s MiB  %14s ms  %14s ms  %16s  %12s ms\n",
+				PQgetvalue(res, i, 0),
+				PQgetvalue(res, i, 1),
+				PQgetvalue(res, i, 2),
+				PQgetvalue(res, i, 3),
+				PQgetvalue(res, i, 4),
+				PQgetvalue(res, i, 5),
+				PQgetvalue(res, i, 6),
+				PQgetvalue(res, i, 7));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
@@ -816,10 +859,10 @@ report_autovacuum_activity(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "/* Autovacuum Activity */\n");
 	fprintf(out, "----------------------------------------\n\n");
 
-	fprintf(out, "/** Basic Statistics (Average) **/\n");
+	fprintf(out, "/** Vacuum Basic Statistics (Average) **/\n");
 	fprintf(out, "-----------------------------------\n");
 	fprintf(out, "%-32s  %8s  %12s  %12s  %12s  %12s  %13s\n",
-		"Table", "Count", "Index Scans", "Removed Rows", "Remain Rows", "Duration", "Duration(Max)");
+		"Table", "Count", "Removed Rows", "Remain Rows", "Index Scans", "Duration", "Duration(Max)");
 	fprintf(out, "--------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_AUTOVACUUM_ACTIVITY, lengthof(params), params);
@@ -839,7 +882,7 @@ report_autovacuum_activity(PGconn *conn, ReportScope *scope, FILE *out)
 
 	if (scope->version >= 90200)
 	{
-		fprintf(out, "/** I/O Statistics (Average) **/\n");
+		fprintf(out, "/** Vacuum I/O Statistics (Average) **/\n");
 		fprintf(out, "-----------------------------------\n");
 		fprintf(out, "%-32s  %10s  %10s  %10s  %13s  %13s\n",
 			"Table", "Page Hit", "Page Miss", "Page Dirty", "Read Rate", "Write Rate");
@@ -859,6 +902,25 @@ report_autovacuum_activity(PGconn *conn, ReportScope *scope, FILE *out)
 		fprintf(out, "\n");
 		PQclear(res);
 	}
+
+	fprintf(out, "/** Analyze Statistics **/\n");
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-32s  %15s  %15s  %15s  %8s\n",
+		"Table", "Duration(Total)", "Duration(Avg)", "Duration(Max)", "Count");
+	fprintf(out, "------------------------------------------------------------------------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_AUTOANALYZE_STATS, lengthof(params), params);
+	for(i = 0; i < PQntuples(res); i++)
+	{
+		fprintf(out, "%-32s  %13s s  %13s s  %13s s  %8s\n",
+			PQgetvalue(res, i, 0),
+			PQgetvalue(res, i, 1),
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3),
+			PQgetvalue(res, i, 4));
+	}
+	fprintf(out, "\n");
+	PQclear(res);
 }
 
 /*
@@ -977,9 +1039,12 @@ report_replication_activity(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "----------------------------------------\n");
 	fprintf(out, "/* Replication Activity */\n");
-	fprintf(out, "----------------------------------------\n");
+	fprintf(out, "----------------------------------------\n\n");
 
-	res = pgut_execute(conn, SQL_SELECT_REPLICATION_ACTIVITY, lengthof(params), params);
+	fprintf(out, "/** Current Replication Status **/\n");
+	fprintf(out, "-----------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_REPLICATION_STATUS, lengthof(params), params);
 	for (i = 0; i < PQntuples(res); i++)
 	{
 		fprintf(out, "User Name             : %s\n", PQgetvalue(res, i, 0));
@@ -995,7 +1060,32 @@ report_replication_activity(PGconn *conn, ReportScope *scope, FILE *out)
 		fprintf(out, "Flush WAL Location    : %s\n", PQgetvalue(res, i, 10));
 		fprintf(out, "Replay WAL Location   : %s\n", PQgetvalue(res, i, 11));
 		fprintf(out, "Sync Priority         : %s\n", PQgetvalue(res, i, 12));
-		fprintf(out, "Sync State            : %s\n", PQgetvalue(res, i, 13));
+		fprintf(out, "Sync State            : %s\n\n", PQgetvalue(res, i, 13));
+	}
+	if (PQntuples(res) == 0)
+		fprintf(out, "\n");
+	PQclear(res);
+
+	fprintf(out, "/** Replication Delays **/\n");
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-16s  %-18s  %17s  %17s \n",
+		"DateTime", "Client", "Flush Delay Size", "Replay Delay Size");
+	fprintf(out, "-----------------------------------------------------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_REPLICATION_DELAYS_SYNC, lengthof(params), params);
+	if (PQntuples(res) == 0)
+	{
+		PQclear(res);
+		res = pgut_execute(conn, SQL_SELECT_REPLICATION_DELAYS_ASYNC, lengthof(params), params);
+	}
+
+	for (i = 0; i < PQntuples(res); i++)
+	{
+		fprintf(out, "%-16s  %-18s  %12s byte  %12s byte\n",
+			PQgetvalue(res, i, 0),
+			PQgetvalue(res, i, 1),
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
@@ -1014,17 +1104,18 @@ report_setting_parameters(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "----------------------------------------\n");
 	fprintf(out, "/* Setting Parameters */\n");
 	fprintf(out, "----------------------------------------\n");
-	fprintf(out, "%-32s  %-32s  %-s\n",
-		"Name", "Setting", "Source");
-	fprintf(out, "-----------------------------------------------------------------------------------------------\n");
+	fprintf(out, "%-32s  %-32s  %-6s  %-s\n",
+		"Name", "Setting", "Unit", "Source");
+	fprintf(out, "-------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_SETTING_PARAMETERS, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-32s  %-32s  %-s\n",
+		fprintf(out, "%-32s  %-32s  %-6s  %-s\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
-			PQgetvalue(res, i, 2));
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
